@@ -27,18 +27,21 @@ static float cursorx = 0.5f, cursory = 0.5f;
 #define SKIN_H 128
 #define SKIN_SCALE 4
 #define INSERT (3*SKIN_SCALE)
+#define MAXCOLUMNS 16
 
 VARP(guiautotab, 6, 16, 40);
 VARP(guiclicktab, 0, 0, 1);
+VARP(guifadein, 0, 1, 1);
 
 struct gui : g3d_gui
 {
     struct list
     {
-        int parent, w, h;
+        int parent, w, h, springs, curspring, column;
     };
 
-    int nextlist;
+    int firstlist, nextlist;
+    int columns[MAXCOLUMNS];
 
     static vector<list> lists;
     static float hitx, hity;
@@ -105,12 +108,7 @@ struct gui : g3d_gui
         if(curdepth != 0) return;
         if(color) tcolor = color;
         tpos++; 
-        if(!name) 
-        {
-            static string title;
-            formatstring(title)("%d", tpos);
-            name = title;
-        }
+        if(!name) name = intstr(tpos); 
         int w = max(text_width(name) - 2*INSERT, 0);
         if(layoutpass) 
         {  
@@ -127,10 +125,7 @@ struct gui : g3d_gui
                 y2 = cury;
             bool hit = tcurrent && windowhit==this && hitx>=x1 && hity>=y1 && hitx<x2 && hity<y2;
             if(hit && (!guiclicktab || mousebuttons&G3D_DOWN)) 
-            {	
                 *tcurrent = tpos; //roll-over to switch tab
-                color = 0xFF0000;
-            }
             
             drawskin(x1-skinx[visible()?2:6]*SKIN_SCALE, y1-skiny[1]*SKIN_SCALE, w, h, visible()?10:19, 9, gui2d ? 1 : 2, light, alpha);
             text_(name, x1 + (skinx[3]-skinx[2])*SKIN_SCALE - (w ? INSERT : INSERT/2), y1 + (skiny[2]-skiny[1])*SKIN_SCALE - INSERT, tcolor, visible());
@@ -152,35 +147,62 @@ struct gui : g3d_gui
             }
             list &l = lists.add();
             l.parent = curlist;
+            l.springs = 0;
+            l.column = -1;
             curlist = lists.length()-1;
             xsize = ysize = 0;
         }
         else
         {
             curlist = nextlist++;
-            xsize = lists[curlist].w;
-            ysize = lists[curlist].h;
+            if(curlist >= lists.length()) // should never get here unless script code doesn't use same amount of lists in layout and render passes
+            {
+                list &l = lists.add();
+                l.parent = curlist;
+                l.springs = 0;
+                l.column = -1;
+                l.w = l.h = 0;
+            }
+            list &l = lists[curlist];
+            l.curspring = 0;
+            if(l.springs > 0)
+            {
+                if(ishorizontal()) xsize = l.w; else ysize = l.h;
+            }
+            else
+            {
+                xsize = l.w;
+                ysize = l.h;
+            }
         }
         curdepth++;	
     }
 
     void poplist()
     {
+        if(!lists.inrange(curlist)) return;
         list &l = lists[curlist];
         if(layoutpass)
         {
             l.w = xsize;
             l.h = ysize;
+            if(l.column >= 0) columns[l.column] = max(columns[l.column], ishorizontal() ? ysize : xsize);
         }
         curlist = l.parent;
         curdepth--;
-        if(curlist>=0)
+        if(lists.inrange(curlist))
         {   
-            xsize = lists[curlist].w;
-            ysize = lists[curlist].h;
-            if(ishorizontal()) cury -= l.h;
-            else curx -= l.w;
-            layout(l.w, l.h);
+            int w = xsize, h = ysize;
+            if(ishorizontal()) cury -= h; else curx -= w;
+            list &p = lists[curlist];
+            xsize = p.w;
+            ysize = p.h;
+            if(!layoutpass && p.springs > 0)
+            {
+                list &s = lists[p.parent];
+                if(ishorizontal()) xsize = s.w; else ysize = s.h;
+            } 
+            layout(w, h);
         }
     }
 
@@ -188,13 +210,40 @@ struct gui : g3d_gui
     int button(const char *text, int color, const char *icon) { autotab(); return button_(text, color, icon, true, false); }
     int title (const char *text, int color, const char *icon) { autotab(); return button_(text, color, icon, false, true); }
 
-    void separator() { autotab(); line_(5); }
-    void progress(float percent) { autotab(); line_(FONTH*2/5, percent); }
+    void separator() { autotab(); line_(FONTH/3); }
+    void progress(float percent) { autotab(); line_((FONTH*4)/5, percent); }
 
     //use to set min size (useful when you have progress bars)
-    void strut(int size) { layout(isvertical() ? size*FONTW : 0, isvertical() ? 0 : size*FONTH); }
+    void strut(float size) { layout(isvertical() ? int(size*FONTW) : 0, isvertical() ? 0 : int(size*FONTH)); }
     //add space between list items
-    void space(int size) { layout(isvertical() ? 0 : size*FONTW, isvertical() ? size*FONTH : 0); }
+    void space(float size) { layout(isvertical() ? 0 : int(size*FONTW), isvertical() ? int(size*FONTH) : 0); }
+
+    void spring(int weight) 
+    { 
+        if(curlist < 0) return;
+        list &l = lists[curlist];
+        if(layoutpass) { if(l.parent >= 0) l.springs += weight; return; }
+        int nextspring = min(l.curspring + weight, l.springs);
+        if(nextspring <= l.curspring) return;
+        if(ishorizontal())
+        {
+            int w = xsize - l.w;
+            layout((w*nextspring)/l.springs - (w*l.curspring)/l.springs, 0);
+        }
+        else
+        {
+            int h = ysize - l.h;
+            layout(0, (h*nextspring)/l.springs - (h*l.curspring)/l.springs);
+        }
+        l.curspring = nextspring;
+    }
+
+    void column(int col)
+    {
+        if(curlist < 0 || !layoutpass || col < 0 || col >= MAXCOLUMNS) return;
+        list &l = lists[curlist];
+        l.column = col;
+    }
 
     int layout(int w, int h)
     {
@@ -236,33 +285,146 @@ struct gui : g3d_gui
         autotab();
         if(scale==0) scale = 1;
         int size = (int)(scale*2*FONTH)-SHADOW;
-        if(visible()) icon_(t, overlaid, false, curx, cury, size, ishit(size+SHADOW, size+SHADOW));
+        if(visible()) icon_(t, overlaid, curx, cury, size, ishit(size+SHADOW, size+SHADOW));
         return layout(size+SHADOW, size+SHADOW);
     }
     
-    int texture(Texture *t, float scale, int rotate, int xoff, int yoff, Texture *glowtex, const vec &glowcolor, Texture *layertex)
+    int texture(VSlot &vslot, float scale, bool overlaid)
     {
         autotab();
         if(scale==0) scale = 1;
         int size = (int)(scale*2*FONTH)-SHADOW;
-        if(t!=notexture && visible()) icon_(t, true, true, curx, cury, size, ishit(size+SHADOW, size+SHADOW), rotate, xoff, yoff, glowtex, glowcolor, layertex);
+        if(visible()) previewslot(vslot, overlaid, curx, cury, size, ishit(size+SHADOW, size+SHADOW));
         return layout(size+SHADOW, size+SHADOW);
     }
 
-    void slider(int &val, int vmin, int vmax, int color, char *label)
+    int playerpreview(int model, int team, int weap, float sizescale, bool overlaid)
+    {
+        autotab();
+        if(sizescale==0) sizescale = 1;
+        int size = (int)(sizescale*2*FONTH)-SHADOW;
+        if(visible())
+        {
+            bool hit = ishit(size+SHADOW, size+SHADOW);
+            float xs = size, ys = size, xi = curx, yi = cury;
+            if(overlaid && hit && actionon)
+            {
+                glDisable(GL_TEXTURE_2D);
+                notextureshader->set();
+                glColor4f(0, 0, 0, 0.75f);
+                rect_(xi+SHADOW, yi+SHADOW, xs, ys, -1);
+                glEnable(GL_TEXTURE_2D);
+                defaultshader->set();
+            }
+            int x1 = int(floor(screen->w*(xi*scale.x+origin.x))), y1 = int(floor(screen->h*(1 - ((yi+ys)*scale.y+origin.y)))),
+                x2 = int(ceil(screen->w*((xi+xs)*scale.x+origin.x))), y2 = int(ceil(screen->h*(1 - (yi*scale.y+origin.y))));
+            glViewport(x1, y1, x2-x1, y2-y1);
+            glScissor(x1, y1, x2-x1, y2-y1);
+            glEnable(GL_SCISSOR_TEST);
+            glDisable(GL_BLEND);
+            modelpreview::start(overlaid);
+            game::renderplayerpreview(model, team, weap);
+            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
+            glEnable(GL_BLEND);
+            modelpreview::end();
+            glDisable(GL_SCISSOR_TEST);
+            glViewport(0, 0, screen->w, screen->h);
+            if(overlaid)
+            {
+                if(hit)
+                {
+                    glDisable(GL_TEXTURE_2D);
+                    notextureshader->set();
+                    glBlendFunc(GL_ZERO, GL_SRC_COLOR);
+                    glColor3f(1, 0.5f, 0.5f);
+                    rect_(xi, yi, xs, ys, -1);
+                    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
+                    glEnable(GL_TEXTURE_2D);
+                    defaultshader->set();
+                }
+                if(!overlaytex) overlaytex = textureload("data/guioverlay.png", 3);
+                glColor3fv(light.v);
+                glBindTexture(GL_TEXTURE_2D, overlaytex->id);
+                rect_(xi, yi, xs, ys, 0);
+            }
+        }
+        return layout(size+SHADOW, size+SHADOW);
+    }
+
+    int modelpreview(const char *name, int anim, float sizescale, bool overlaid)
+    {
+        autotab();
+        if(sizescale==0) sizescale = 1;
+        int size = (int)(sizescale*2*FONTH)-SHADOW;
+        if(visible())
+        {
+            bool hit = ishit(size+SHADOW, size+SHADOW);
+            float xs = size, ys = size, xi = curx, yi = cury;
+            if(overlaid && hit && actionon)
+            {
+                glDisable(GL_TEXTURE_2D);
+                notextureshader->set();
+                glColor4f(0, 0, 0, 0.75f);
+                rect_(xi+SHADOW, yi+SHADOW, xs, ys, -1);
+                glEnable(GL_TEXTURE_2D);
+                defaultshader->set();
+            }
+            int x1 = int(floor(screen->w*(xi*scale.x+origin.x))), y1 = int(floor(screen->h*(1 - ((yi+ys)*scale.y+origin.y)))),
+                x2 = int(ceil(screen->w*((xi+xs)*scale.x+origin.x))), y2 = int(ceil(screen->h*(1 - (yi*scale.y+origin.y))));
+            glViewport(x1, y1, x2-x1, y2-y1);
+            glScissor(x1, y1, x2-x1, y2-y1);
+            glEnable(GL_SCISSOR_TEST);
+            glDisable(GL_BLEND);
+            modelpreview::start(overlaid);
+            model *m = loadmodel(name);
+            if(m)
+            {
+                entitylight light;
+                light.color = vec(1, 1, 1);
+                light.dir = vec(0, -1, 2).normalize();
+                vec center, radius;
+                m->boundbox(0, center, radius);
+                float dist =  2.0f*max(radius.magnitude2(), 1.1f*radius.z),
+                      yaw = fmod(lastmillis/10000.0f*360.0f, 360.0f);
+                vec o(-center.x, dist - center.y, -0.1f*dist - center.z);
+                rendermodel(&light, name, anim, o, yaw, 0, 0, NULL, NULL, 0);
+            }
+            modelpreview::end();
+            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
+            glEnable(GL_BLEND);
+            glDisable(GL_SCISSOR_TEST);
+            glViewport(0, 0, screen->w, screen->h);
+            if(overlaid)
+            {
+                if(hit)
+                {
+                    glDisable(GL_TEXTURE_2D);
+                    notextureshader->set();
+                    glBlendFunc(GL_ZERO, GL_SRC_COLOR);
+                    glColor3f(1, 0.5f, 0.5f);
+                    rect_(xi, yi, xs, ys, -1);
+                    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
+                    glEnable(GL_TEXTURE_2D);
+                    defaultshader->set();
+                }
+                if(!overlaytex) overlaytex = textureload("data/guioverlay.png", 3);
+                glColor3fv(light.v);
+                glBindTexture(GL_TEXTURE_2D, overlaytex->id);
+                rect_(xi, yi, xs, ys, 0);
+            }
+        }
+        return layout(size+SHADOW, size+SHADOW);
+    }
+
+    void slider(int &val, int vmin, int vmax, int color, const char *label)
     {
         autotab();
         int x = curx;
         int y = cury;
-        line_(10);
+        line_((FONTH*2)/3);
         if(visible())
         {
-            if(!label)
-            {
-                static string s;
-                formatstring(s)("%d", val);
-                label = s;
-            }
+            if(!label) label = intstr(val);
             int w = text_width(label);
 
             bool hit;
@@ -281,12 +443,12 @@ struct gui : g3d_gui
             }
 
             if(hit) color = 0xFF0000;
-            text_(label, px, py, color, hit && actionon);
+            text_(label, px, py, color, hit && actionon, hit);
             if(hit && actionon)
             {
                 int vnew = (vmin < vmax ? 1 : -1)+vmax-vmin;
-                if(ishorizontal()) vnew = int(vnew*(y+ysize-FONTH/2-hity)/(ysize-FONTH));
-                else vnew = int(vnew*(hitx-x-FONTH/2)/(xsize-w));
+                if(ishorizontal()) vnew = int((vnew*(y+ysize-FONTH/2-hity))/(ysize-FONTH));
+                else vnew = int((vnew*(hitx-x-FONTH/2))/(xsize-w));
                 vnew += vmin;
                 vnew = vmin < vmax ? clamp(vnew, vmin, vmax) : clamp(vnew, vmax, vmin);
                 if(vnew != val) val = vnew;
@@ -334,6 +496,8 @@ struct gui : g3d_gui
         char *result = NULL;
         if(visible() && !layoutpass)
         {
+            e->rendered = true;
+
             bool hit = ishit(w, h);
             if(hit) 
             {
@@ -357,21 +521,21 @@ struct gui : g3d_gui
             
             e->draw(curx+FONTW/2, cury, color, hit && editing);
             
-            notextureshader->set();
+            lineshader->set();
             glDisable(GL_TEXTURE_2D);
+            glDisable(GL_BLEND);
             if(editing) glColor3f(1, 0, 0);
             else glColor3ub(color>>16, (color>>8)&0xFF, color&0xFF);
-            glBegin(GL_LINE_LOOP);
-            rect_(curx, cury, w, h);
-            glEnd();
+            rect_(curx, cury, w, h, -1, true);
             glEnable(GL_TEXTURE_2D);
+            glEnable(GL_BLEND);
             defaultshader->set();
         }
         layout(w, h);
         
         if(e->maxy != 1)
         {
-            int slines = e->lines.length()-e->pixelheight/FONTH;
+            int slines = e->limitscrolly();
             if(slines > 0) 
             {
                 int pos = e->scrolly;
@@ -384,24 +548,35 @@ struct gui : g3d_gui
         return result;
     }
 
-    void rect_(float x, float y, float w, float h, int usetc = -1) 
+    void rect_(float x, float y, float w, float h, int usetc = -1, bool lines = false) 
     {
+        glBegin(lines ? GL_LINE_LOOP : GL_TRIANGLE_STRIP);
         static const GLfloat tc[4][2] = {{0, 0}, {1, 0}, {1, 1}, {0, 1}};
         if(usetc>=0) glTexCoord2fv(tc[usetc]); 
         glVertex2f(x, y);
         if(usetc>=0) glTexCoord2fv(tc[(usetc+1)%4]);
         glVertex2f(x + w, y);
-        if(usetc>=0) glTexCoord2fv(tc[(usetc+2)%4]);
-        glVertex2f(x + w, y + h);
+        if(lines)
+        {
+            if(usetc>=0) glTexCoord2fv(tc[(usetc+2)%4]);
+            glVertex2f(x + w, y + h);
+        }
         if(usetc>=0) glTexCoord2fv(tc[(usetc+3)%4]);
         glVertex2f(x, y + h);
+        if(!lines)
+        {
+            if(usetc>=0) glTexCoord2fv(tc[(usetc+2)%4]);
+            glVertex2f(x + w, y + h);
+        }
+        glEnd();
         xtraverts += 4;
+        
     }
 
-    void text_(const char *text, int x, int y, int color, bool shadow) 
+    void text_(const char *text, int x, int y, int color, bool shadow, bool force = false) 
     {
-        if(shadow) draw_text(text, x+SHADOW, y+SHADOW, 0x00, 0x00, 0x00, 0xC0);
-        draw_text(text, x, y, color>>16, (color>>8)&0xFF, color&0xFF);
+        if(shadow) draw_text(text, x+SHADOW, y+SHADOW, 0x00, 0x00, 0x00, -0xC0);
+        draw_text(text, x, y, color>>16, (color>>8)&0xFF, color&0xFF, force ? -0xFF : 0xFF);
     }
 
     void background(int color, int inheritw, int inherith)
@@ -413,120 +588,133 @@ struct gui : g3d_gui
         int w = xsize, h = ysize;
         if(inheritw>0) 
         {
-            int parentw = curlist;
-            while(inheritw>0 && lists[parentw].parent>=0)
-            {
+            int parentw = curlist, parentdepth = 0;
+            for(;parentdepth < inheritw && lists[parentw].parent>=0; parentdepth++)
                 parentw = lists[parentw].parent;
-                inheritw--;
-            }
-            w = lists[parentw].w;
+            list &p = lists[parentw];
+            w = p.springs > 0 && (curdepth-parentdepth)&1 ? lists[p.parent].w : p.w;
         }
         if(inherith>0)
         {
-            int parenth = curlist;
-            while(inherith>0 && lists[parenth].parent>=0)
-            {
+            int parenth = curlist, parentdepth = 0;
+            for(;parentdepth < inherith && lists[parenth].parent>=0; parentdepth++)
                 parenth = lists[parenth].parent;
-                inherith--;
-            }
-            h = lists[parenth].h;
+            list &p = lists[parenth];
+            h = p.springs > 0 && !((curdepth-parentdepth)&1) ? lists[p.parent].h : p.h;
         }
-        glBegin(GL_QUADS);
         rect_(curx, cury, w, h);
-        glEnd();
         glEnable(GL_TEXTURE_2D);
         defaultshader->set();
     }
 
-    void icon_(Texture *t, bool overlaid, bool tiled, int x, int y, int size, bool hit, int rotate = 0, int xoff = 0, int yoff = 0, Texture *glowtex = NULL, const vec &glowcolor = vec(1, 1, 1), Texture *layertex = NULL) 
+    void icon_(Texture *t, bool overlaid, int x, int y, int size, bool hit)
     {
-        float xs, ys, xt, yt;
-        if(tiled)
-        {   
-            xt = min(1.0f, t->xs/(float)t->ys), 
-            yt = min(1.0f, t->ys/(float)t->xs);
-            xs = size;
-            ys = size;
-        } 
-        else 
+        float scale = float(size)/max(t->xs, t->ys); //scale and preserve aspect ratio
+        float xs = t->xs*scale, ys = t->ys*scale;
+        x += int((size-xs)/2);
+        y += int((size-ys)/2);
+        const vec &color = hit ? vec(1, 0.5f, 0.5f) : (overlaid ? vec(1, 1, 1) : light);
+        glBindTexture(GL_TEXTURE_2D, t->id);
+        if(hit && actionon)
         {
-            xt = 1.0f;
-            yt = 1.0f;
-            float scale = float(size)/max(t->xs, t->ys); //scale and preserve aspect ratio
-            xs = t->xs*scale;
-            ys = t->ys*scale;
-            x += int((size-xs)/2);
-            y += int((size-ys)/2);
+            glColor4f(0, 0, 0, 0.75f);
+            rect_(x+SHADOW, y+SHADOW, xs, ys, 0);
         }
+        glColor3fv(color.v);
+        rect_(x, y, xs, ys, 0);
+
+        if(overlaid)
+        {
+            if(!overlaytex) overlaytex = textureload("data/guioverlay.png", 3);
+            glBindTexture(GL_TEXTURE_2D, overlaytex->id);
+            glColor3fv(light.v);
+            rect_(x, y, xs, ys, 0);
+        }
+    }        
+
+    void previewslot(VSlot &vslot, bool overlaid, int x, int y, int size, bool hit)
+    {
+        Slot &slot = *vslot.slot;
+        if(slot.sts.empty()) return;
+        VSlot *layer = NULL;
+        Texture *t = NULL, *glowtex = NULL, *layertex = NULL;
+        if(slot.loaded)
+        {
+            t = slot.sts[0].t;
+            if(t == notexture) return;
+            Slot &slot = *vslot.slot;
+            if(slot.texmask&(1<<TEX_GLOW)) { loopvj(slot.sts) if(slot.sts[j].type==TEX_GLOW) { glowtex = slot.sts[j].t; break; } }
+            if(vslot.layer)
+            {
+                layer = &lookupvslot(vslot.layer);
+                if(!layer->slot->sts.empty()) layertex = layer->slot->sts[0].t;
+            }
+        }
+        else if(slot.thumbnail && slot.thumbnail != notexture) t = slot.thumbnail;
+        else return;
+        float xt = min(1.0f, t->xs/(float)t->ys), yt = min(1.0f, t->ys/(float)t->xs), xs = size, ys = size;
         if(hit && actionon) 
         {
             glDisable(GL_TEXTURE_2D);
             notextureshader->set();
             glColor4f(0, 0, 0, 0.75f);
-            glBegin(GL_QUADS);
             rect_(x+SHADOW, y+SHADOW, xs, ys);
-            glEnd();
             glEnable(GL_TEXTURE_2D);
             defaultshader->set();	
         }
-        if(tiled) 
-        {
-            static Shader *rgbonlyshader = NULL;
-            if(!rgbonlyshader) rgbonlyshader = lookupshaderbyname("rgbonly");
-            rgbonlyshader->set();
-        }
+        SETSHADER(rgbonly);
+        const vec &color = hit ? vec(1, 0.5f, 0.5f) : (overlaid ? vec(1, 1, 1) : light);
         float tc[4][2] = { { 0, 0 }, { 1, 0 }, { 1, 1 }, { 0, 1 } };
-        if(rotate)
+        int xoff = vslot.xoffset, yoff = vslot.yoffset;
+        if(vslot.rotation)
         {
-            if((rotate&5) == 1) { swap(xoff, yoff); loopk(4) swap(tc[k][0], tc[k][1]); }
-            if(rotate >= 2 && rotate <= 4) { xoff *= -1; loopk(4) tc[k][0] *= -1; }
-            if(rotate <= 2 || rotate == 5) { yoff *= -1; loopk(4) tc[k][1] *= -1; }
+            if((vslot.rotation&5) == 1) { swap(xoff, yoff); loopk(4) swap(tc[k][0], tc[k][1]); }
+            if(vslot.rotation >= 2 && vslot.rotation <= 4) { xoff *= -1; loopk(4) tc[k][0] *= -1; }
+            if(vslot.rotation <= 2 || vslot.rotation == 5) { yoff *= -1; loopk(4) tc[k][1] *= -1; }
         }
         loopk(4) { tc[k][0] = tc[k][0]/xt - float(xoff)/t->xs; tc[k][1] = tc[k][1]/yt - float(yoff)/t->ys; }
-        vec color = hit ? vec(1, 0.5f, 0.5f) : (overlaid ? vec(1, 1, 1) : light);
+        if(slot.loaded) glColor3f(color.x*vslot.colorscale.x, color.y*vslot.colorscale.y, color.z*vslot.colorscale.z);
+        else glColor3fv(color.v);
         glBindTexture(GL_TEXTURE_2D, t->id);
-        glColor3fv(color.v);
-        glBegin(GL_QUADS);
+        glBegin(GL_TRIANGLE_STRIP);
         glTexCoord2fv(tc[0]); glVertex2f(x,    y);
         glTexCoord2fv(tc[1]); glVertex2f(x+xs, y);
-        glTexCoord2fv(tc[2]); glVertex2f(x+xs, y+ys);
         glTexCoord2fv(tc[3]); glVertex2f(x,    y+ys);
+        glTexCoord2fv(tc[2]); glVertex2f(x+xs, y+ys);
         glEnd();
         if(glowtex)
         {
             glBlendFunc(GL_SRC_ALPHA, GL_ONE);
             glBindTexture(GL_TEXTURE_2D, glowtex->id);
-            if(hit || overlaid) { loopk(3) color[k] *= glowcolor[k]; glColor3fv(color.v); }
-            else glColor3fv(glowcolor.v);
-            glBegin(GL_QUADS);
+            if(hit || overlaid) glColor3f(color.x*vslot.glowcolor.x, color.y*vslot.glowcolor.y, color.z*vslot.glowcolor.z);
+            else glColor3fv(vslot.glowcolor.v);
+            glBegin(GL_TRIANGLE_STRIP);
             glTexCoord2fv(tc[0]); glVertex2f(x,    y);
             glTexCoord2fv(tc[1]); glVertex2f(x+xs, y);
-            glTexCoord2fv(tc[2]); glVertex2f(x+xs, y+ys);
             glTexCoord2fv(tc[3]); glVertex2f(x,    y+ys);
+            glTexCoord2fv(tc[2]); glVertex2f(x+xs, y+ys);
             glEnd();
             glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
         }
         if(layertex)
         {
             glBindTexture(GL_TEXTURE_2D, layertex->id);
-            glColor3fv(color.v);
-            glBegin(GL_QUADS);
+            glColor3f(color.x*layer->colorscale.x, color.y*layer->colorscale.y, color.z*layer->colorscale.z);
+            glBegin(GL_TRIANGLE_STRIP);
             glTexCoord2fv(tc[0]); glVertex2f(x+xs/2, y+ys/2);
             glTexCoord2fv(tc[1]); glVertex2f(x+xs,   y+ys/2);
-            glTexCoord2fv(tc[2]); glVertex2f(x+xs,   y+ys);
             glTexCoord2fv(tc[3]); glVertex2f(x+xs/2, y+ys);
+            glTexCoord2fv(tc[2]); glVertex2f(x+xs,   y+ys);
             glEnd();
         }
             
-        if(tiled) defaultshader->set();
+        defaultshader->set();
         if(overlaid) 
         {
             if(!overlaytex) overlaytex = textureload("data/guioverlay.png", 3);
             glBindTexture(GL_TEXTURE_2D, overlaytex->id);
             glColor3fv(light.v);
-            glBegin(GL_QUADS);
             rect_(x, y, xs, ys, 0);
-            glEnd();
         }
     }
 
@@ -535,23 +723,20 @@ struct gui : g3d_gui
         if(visible())
         {
             if(!slidertex) slidertex = textureload("data/guislider.png", 3);
-            glEnable(GL_TEXTURE_2D);
             glBindTexture(GL_TEXTURE_2D, slidertex->id);
-            glBegin(GL_QUADS);
             if(percent < 0.99f) 
             {
                 glColor4f(light.x, light.y, light.z, 0.375f);
                 if(ishorizontal()) 
-                    rect_(curx + FONTH/2 - size, cury, size*2, ysize, 0);
+                    rect_(curx + FONTH/2 - size/2, cury, size, ysize, 0);
                 else
-                    rect_(curx, cury + FONTH/2 - size, xsize, size*2, 1);
+                    rect_(curx, cury + FONTH/2 - size/2, xsize, size, 1);
             }
             glColor3fv(light.v);
             if(ishorizontal()) 
-                rect_(curx + FONTH/2 - size, cury + ysize*(1-percent), size*2, ysize*percent, 0);
+                rect_(curx + FONTH/2 - size/2, cury + ysize*(1-percent), size, ysize*percent, 0);
             else 
-                rect_(curx, cury + FONTH/2 - size, xsize*percent, size*2, 1);
-            glEnd();
+                rect_(curx, cury + FONTH/2 - size/2, xsize*percent, size, 1);
         }
         layout(ishorizontal() ? FONTH : 0, ishorizontal() ? 0 : FONTH);
     }
@@ -584,12 +769,16 @@ struct gui : g3d_gui
         
             if(icon)
             {
-                defformatstring(tname)("packages/icons/%s.jpg", icon);
-                icon_(textureload(tname, 3), false, false, x, cury, ICON_SIZE, clickable && hit);
+                if(icon[0] != ' ')
+                {
+                    const char *ext = strrchr(icon, '.');
+                    defformatstring(tname)("packages/icons/%s%s", icon, ext ? "" : ".jpg");
+                    icon_(textureload(tname, 3), false, x, cury, ICON_SIZE, clickable && hit);
+                }
                 x += ICON_SIZE;
             }
             if(icon && text) x += padding;
-            if(text) text_(text, x, cury, color, center || (hit && clickable && actionon));
+            if(text) text_(text, x, cury, color, center || (hit && clickable && actionon), hit && clickable);
         }
         return layout(w, FONTH);
     }
@@ -694,7 +883,7 @@ struct gui : g3d_gui
         if(tcurrent) h += ((skiny[5]-skiny[1])-(skiny[3]-skiny[2]))*SKIN_SCALE + FONTH-2*INSERT;
         else h += (skiny[6]-skiny[3])*SKIN_SCALE;
 
-        float aspect = float(screen->h)/float(screen->w), fit = 1.0f;
+        float aspect = forceaspect ? 1.0f/forceaspect : float(screen->h)/float(screen->w), fit = 1.0f;
         if(w*aspect*basescale>1.0f) fit = 1.0f/(w*aspect*basescale);
         if(h*basescale*fit>maxscale) fit *= maxscale/(h*basescale*fit);
         origin = vec(0.5f-((w-xsize)/2 - (skinx[2]-skinx[1])*SKIN_SCALE)*aspect*scale.x*fit, 0.5f + (0.5f*h-(skiny[9]-skiny[7])*SKIN_SCALE)*scale.y*fit, 0);
@@ -709,7 +898,7 @@ struct gui : g3d_gui
             if(allowinput) hascursor = true;
         }
         basescale = initscale;
-        if(layoutpass) scale.x = scale.y = scale.z = min(basescale*(totalmillis-starttime)/300.0f, basescale);
+        if(layoutpass) scale.x = scale.y = scale.z = guifadein ? basescale*min((totalmillis-starttime)/300.0f, 1.0f) : basescale;
         alpha = allowinput ? 0.80f : 0.60f;
         passthrough = scale.x<basescale || !allowinput;
         curdepth = -1;
@@ -720,7 +909,11 @@ struct gui : g3d_gui
         tcurrent = tab;
         tcolor = 0xFFFFFF;
         pushlist();
-        if(layoutpass) nextlist = curlist;
+        if(layoutpass) 
+        {
+            firstlist = nextlist = curlist;
+            memset(columns, 0, sizeof(columns));
+        }
         else
         {
             if(tcurrent && !*tcurrent) tcurrent = NULL;
@@ -743,7 +936,7 @@ struct gui : g3d_gui
                 glScalef(-scale.x, scale.y, scale.z);
             
                 vec dir;
-                lightreaching(origin, light, dir, 0, 0.5f); 
+                lightreaching(origin, light, dir, false, 0, 0.5f); 
                 float intensity = vec(yaw, 0.0f).dot(dir);
                 light.mul(1.0f + max(intensity, 0.0f));
             }
@@ -753,10 +946,58 @@ struct gui : g3d_gui
         }
     }
 
+    void adjusthorizontalcolumn(int col, int i)
+    {
+        int h = columns[col], dh = 0;
+        for(int d = 1; i >= 0; d ^= 1)
+        {
+            list &p = lists[i];
+            if(d&1) { dh = h - p.h; if(dh <= 0) break; p.h = h; }
+            else { p.h += dh; h = p.h; }
+            i = p.parent;
+        }
+        ysize += max(dh, 0);
+    }
+
+    void adjustverticalcolumn(int col, int i)
+    {
+        int w = columns[col], dw = 0;
+        for(int d = 0; i >= 0; d ^= 1)
+        {
+            list &p = lists[i];
+            if(d&1) { p.w += dw; w = p.w; }
+            else { dw = w - p.w; if(dw <= 0) break; p.w = w; }
+            i = p.parent;
+        }
+        xsize = max(xsize, w);
+    }
+        
+    void adjustcolumns()
+    {
+        if(lists.inrange(curlist))
+        {
+            list &l = lists[curlist];
+            if(l.column >= 0) columns[l.column] = max(columns[l.column], ishorizontal() ? ysize : xsize);
+        }
+        int parent = -1, depth = 0;
+        for(int i = firstlist; i < lists.length(); i++)
+        {
+            list &l = lists[i];
+            if(l.parent > parent) { parent = l.parent; depth++; }
+            else if(l.parent < parent) { parent = l.parent; depth--; }
+            if(l.column >= 0)
+            {
+                if(depth&1) adjusthorizontalcolumn(l.column, i); 
+                else adjustverticalcolumn(l.column, i);
+            }
+        }
+    }
+
     void end()
     {
         if(layoutpass)
         {	
+            adjustcolumns();
             xsize = max(tx, xsize);
             ysize = max(ty, ysize);
             ysize = max(ysize, (skiny[7]-skiny[6])*SKIN_SCALE);
@@ -797,6 +1038,11 @@ struct gui : g3d_gui
         }
         poplist();
     }
+
+    void draw()
+    {
+        cb->gui(*this, layoutpass);
+    }
 };
 
 Texture *gui::skintex = NULL, *gui::overlaytex = NULL, *gui::slidertex = NULL;
@@ -947,6 +1193,7 @@ bool menukey(int code, bool isdown, int cooked)
             break;
         default:
             if(!cooked || (code<32)) return false;
+            break;
     }
     if(!isdown) return true;
     e->key(code, cooked);
@@ -964,12 +1211,14 @@ void g3d_resetcursor()
     cursorx = cursory = 0.5f;
 }
 
+FVARP(guisens, 1e-3f, 1, 1e3f);
+
 bool g3d_movecursor(int dx, int dy)
 {
     if(!guis2d.length() || !hascursor) return false;
     const float CURSORSCALE = 500.0f;
-    cursorx = max(0.0f, min(1.0f, cursorx+dx*(screen->h/(screen->w*CURSORSCALE))));
-    cursory = max(0.0f, min(1.0f, cursory+dy/CURSORSCALE));
+    cursorx = max(0.0f, min(1.0f, cursorx+guisens*dx*(screen->h/(screen->w*CURSORSCALE))));
+    cursory = max(0.0f, min(1.0f, cursory+guisens*dy/CURSORSCALE));
     return true;
 }
 
@@ -984,7 +1233,7 @@ void g3d_addgui(g3d_callback *cb, vec &origin, int flags)
     g.cb = cb;
     g.origin = origin;
     g.savedorigin = &origin;
-    g.dist = camera1->o.dist(g.origin);
+    g.dist = flags&GUI_BOTTOM && gui2d ? 1e16f : camera1->o.dist(g.origin);
     g.gui2d = gui2d;
 }
 
@@ -993,7 +1242,7 @@ void g3d_limitscale(float scale)
     gui::maxscale = scale;
 }
 
-int g3d_sort(gui *a, gui *b) { return (int)(a->dist>b->dist)*2-1; }
+static inline bool g3d_sort(const gui &a, const gui &b) { return a.dist < b.dist; }
 
 bool g3d_windowhit(bool on, bool act)
 {
@@ -1013,10 +1262,10 @@ void g3d_render()
 {
     windowhit = NULL;    
     if(actionon) mousebuttons |= G3D_PRESSED;
-    
-    gui::reset();
-    guis2d.setsize(0);
-    guis3d.setsize(0);
+   
+    gui::reset(); 
+    guis2d.shrink(0);
+    guis3d.shrink(0);
  
     // call all places in the engine that may want to render a gui from here, they call g3d_addgui()
     extern void g3d_texturemenu();
@@ -1035,8 +1284,8 @@ void g3d_render()
     hascursor = false;
 
     layoutpass = true;
-    loopv(guis2d) guis2d[i].cb->gui(guis2d[i], true);
-    loopv(guis3d) guis3d[i].cb->gui(guis3d[i], true);
+    loopv(guis2d) guis2d[i].draw();
+    loopv(guis3d) guis3d[i].draw();
     layoutpass = false;
 
     if(guis2d.length() || guis3d.length())
@@ -1051,7 +1300,7 @@ void g3d_render()
         glDepthFunc(GL_ALWAYS);
         glDepthMask(GL_FALSE);
 
-        loopvrev(guis3d) guis3d[i].cb->gui(guis3d[i], false);
+        loopvrev(guis3d) guis3d[i].draw();
 
         glDepthFunc(GL_LESS);
         glDepthMask(GL_TRUE);
@@ -1069,7 +1318,7 @@ void g3d_render()
         glPushMatrix();
         glLoadIdentity();
 
-        loopvrev(guis2d) guis2d[i].cb->gui(guis2d[i], false);
+        loopvrev(guis2d) guis2d[i].draw();
 
         glMatrixMode(GL_PROJECTION);
         glPopMatrix();
