@@ -16,10 +16,10 @@ bool multiplayer(bool msg)
 void setrate(int rate)
 {
    if(!curpeer) return;
-   enet_host_bandwidth_limit(clienthost, rate, rate);
+   enet_host_bandwidth_limit(clienthost, rate*1024, rate*1024);
 }
 
-VARF(rate, 0, 0, 25000, setrate(rate));
+VARF(rate, 0, 0, 1024, setrate(rate));
 
 void throttle();
 
@@ -34,12 +34,30 @@ void throttle()
     enet_peer_throttle_configure(curpeer, throttle_interval*1000, throttle_accel, throttle_decel);
 }
 
-bool isconnected(bool attempt)
+bool isconnected(bool attempt, bool local)
 {
-    return curpeer || (attempt && connpeer);
+    return curpeer || (attempt && connpeer) || (local && haslocalclients());
 }
 
-ICOMMAND(isconnected, "i", (int *attempt), intret(isconnected(*attempt > 0) ? 1 : 0));
+ICOMMAND(isconnected, "bb", (int *attempt, int *local), intret(isconnected(*attempt > 0, *local != 0) ? 1 : 0));
+
+const ENetAddress *connectedpeer()
+{
+    return curpeer ? &curpeer->address : NULL;
+}
+
+ICOMMAND(connectedip, "", (),
+{
+    const ENetAddress *address = connectedpeer();
+    string hostname;
+    result(address && enet_address_get_host_ip(address, hostname, sizeof(hostname)) >= 0 ? hostname : "");
+});
+
+ICOMMAND(connectedport, "", (),
+{
+    const ENetAddress *address = connectedpeer();
+    intret(address ? address->port : -1);
+});
 
 void abortconnect()
 {
@@ -52,6 +70,9 @@ void abortconnect()
     clienthost = NULL;
 }
 
+SVARP(connectname, "");
+VARP(connectport, 0, 0, 0xFFFF);
+
 void connectserv(const char *servername, int serverport, const char *serverpassword)
 {   
     if(connpeer)
@@ -67,7 +88,9 @@ void connectserv(const char *servername, int serverport, const char *serverpassw
 
     if(servername)
     {
-        addserver(servername, serverport, serverpassword[0] ? serverpassword : NULL);
+        if(strcmp(servername, connectname)) setsvar("connectname", servername);
+        if(serverport != connectport) setvar("connectport", serverport);
+        addserver(servername, serverport, serverpassword && serverpassword[0] ? serverpassword : NULL);
         conoutf("attempting to connect to %s:%d", servername, serverport);
         if(!resolverwait(servername, &address))
         {
@@ -77,15 +100,18 @@ void connectserv(const char *servername, int serverport, const char *serverpassw
     }
     else
     {
+        setsvar("connectname", "");
+        setvar("connectport", 0);
         conoutf("attempting to connect over LAN");
         address.host = ENET_HOST_BROADCAST;
     }
 
-    if(!clienthost) clienthost = enet_host_create(NULL, 2, rate, rate);
+    if(!clienthost) 
+        clienthost = enet_host_create(NULL, 2, server::numchannels(), rate*1024, rate*1024);
 
     if(clienthost)
     {
-        connpeer = enet_host_connect(clienthost, &address, game::numchannels()); 
+        connpeer = enet_host_connect(clienthost, &address, server::numchannels(), 0); 
         enet_host_flush(clienthost);
         connmillis = totalmillis;
         connattempts = 0;
@@ -95,6 +121,17 @@ void connectserv(const char *servername, int serverport, const char *serverpassw
     else conoutf("\f3could not connect to server");
 }
 
+void reconnect(const char *serverpassword)
+{
+    if(!connectname[0] || connectport <= 0)
+    {
+        conoutf(CON_ERROR, "no previous connection");
+        return;
+    }
+
+    connectserv(connectname, connectport, serverpassword);
+}
+
 void disconnect(bool async, bool cleanup)
 {
     if(curpeer) 
@@ -123,7 +160,7 @@ void disconnect(bool async, bool cleanup)
     }
 }
 
-void trydisconnect()
+void trydisconnect(bool local)
 {
     if(connpeer)
     {
@@ -135,13 +172,15 @@ void trydisconnect()
         conoutf("attempting to disconnect...");
         disconnect(!discmillis);
     }
+    else if(local && haslocalclients()) localdisconnect();
     else conoutf("not connected");
 }
 
 ICOMMAND(connect, "sis", (char *name, int *port, char *pw), connectserv(name, *port, pw));
 ICOMMAND(lanconnect, "is", (int *port, char *pw), connectserv(NULL, *port, pw));
-COMMANDN(disconnect, trydisconnect, "");
-ICOMMAND(localconnect, "", (), { if(!isconnected() && !haslocalclients()) localconnect(); });
+COMMAND(reconnect, "s");
+ICOMMAND(disconnect, "b", (int *local), trydisconnect(*local != 0));
+ICOMMAND(localconnect, "", (), { if(!isconnected()) localconnect(); });
 ICOMMAND(localdisconnect, "", (), { if(haslocalclients()) localdisconnect(); });
 
 void sendclientpacket(ENetPacket *packet, int chan)
@@ -206,7 +245,6 @@ void gets2c()           // get updates from the server
             break;
 
         case ENET_EVENT_TYPE_DISCONNECT:
-            extern const char *disc_reasons[];
             if(event.data>=DISC_NUM) event.data = DISC_NONE;
             if(event.peer==connpeer)
             {
@@ -215,7 +253,12 @@ void gets2c()           // get updates from the server
             }
             else
             {
-                if(!discmillis || event.data) conoutf("\f3server network error, disconnecting (%s) ...", disc_reasons[event.data]);
+                if(!discmillis || event.data)
+                {
+                    const char *msg = disconnectreason(event.data);
+                    if(msg) conoutf("\f3server network error, disconnecting (%s) ...", msg);
+                    else conoutf("\f3server network error, disconnecting...");
+                }
                 disconnect();
             }
             return;
