@@ -13,15 +13,17 @@ static g3d_gui *cgui = NULL;
 
 struct menu : g3d_callback
 {
-    char *name, *header, *contents, *onclear;
+    char *name, *header;
+    uint *contents, *init, *onclear;
+    bool showtab;
 
-    menu() : name(NULL), header(NULL), contents(NULL), onclear(NULL) {}
+    menu() : name(NULL), header(NULL), contents(NULL), init(NULL), onclear(NULL), showtab(true) {}
 
     void gui(g3d_gui &g, bool firstpass)
     {
         cgui = &g;
-        cgui->start(menustart, 0.03f, &menutab);
-        cgui->tab(header ? header : name, GUI_TITLE_COLOR);
+        cgui->start(menustart, 0.03f, showtab ? &menutab : NULL);
+        if(showtab) cgui->tab(header ? header : name, GUI_TITLE_COLOR);
         execute(contents);
         cgui->end();
         cgui = NULL;
@@ -29,13 +31,83 @@ struct menu : g3d_callback
 
     virtual void clear() 
     {
-        DELETEA(onclear);
+        if(onclear) { freecode(onclear); onclear = NULL; }
     }
 };
 
+struct delayedupdate
+{
+    enum
+    {
+        INT,
+        FLOAT,
+        STRING,
+        ACTION
+    } type;
+    ident *id;
+    union
+    {
+        int i;
+        float f;
+        char *s;
+    } val;
+    delayedupdate() : type(ACTION), id(NULL) { val.s = NULL; }
+    ~delayedupdate() { if(type == STRING || type == ACTION) DELETEA(val.s); }
+
+    void schedule(const char *s) { type = ACTION; val.s = newstring(s); }
+    void schedule(ident *var, int i) { type = INT; id = var; val.i = i; }
+    void schedule(ident *var, float f) { type = FLOAT; id = var; val.f = f; }
+    void schedule(ident *var, char *s) { type = STRING; id = var; val.s = newstring(s); }
+
+    int getint() const
+    {
+        switch(type)
+        {
+            case INT: return val.i;
+            case FLOAT: return int(val.f);
+            case STRING: return int(strtol(val.s, NULL, 0));
+            default: return 0;
+        }
+    }
+
+    float getfloat() const
+    {
+        switch(type)
+        {
+            case INT: return float(val.i);
+            case FLOAT: return val.f;
+            case STRING: return float(parsefloat(val.s));
+            default: return 0;
+        }
+    }
+   
+    const char *getstring() const
+    {
+        switch(type)
+        {
+            case INT: return intstr(val.i);
+            case FLOAT: return intstr(int(floor(val.f)));
+            case STRING: return val.s;
+            default: return "";
+        }
+    }
+
+    void run()
+    {
+        if(type == ACTION) { if(val.s) execute(val.s); }
+        else if(id) switch(id->type)
+        {
+            case ID_VAR: setvarchecked(id, getint()); break;
+            case ID_FVAR: setfvarchecked(id, getfloat()); break;
+            case ID_SVAR: setsvarchecked(id, getstring()); break;
+            case ID_ALIAS: alias(id->name, getstring()); break;
+        }
+    }
+};
+     
 static hashtable<const char *, menu> guis;
 static vector<menu *> guistack;
-static vector<char *> executelater;
+static vector<delayedupdate> updatelater;
 static bool shouldclearmenu = true, clearlater = false;
 
 VARP(menudistance,  16, 40,  256);
@@ -80,6 +152,7 @@ void pushgui(menu *m, int pos = -1)
         menutab = 1;
         menustart = totalmillis;
     }
+    if(m->init) execute(m->init);
 }
 
 void restoregui(int pos)
@@ -99,6 +172,12 @@ void showgui(const char *name)
     else restoregui(pos);
 }
 
+void hidegui(const char *name)
+{
+    menu *m = guis.access(name);
+    if(m) removegui(m);
+}
+ 
 int cleargui(int n)
 {
     int clear = guistack.length();
@@ -113,15 +192,32 @@ int cleargui(int n)
     return clear;
 }
 
+void clearguis(int level = -1)
+{
+    if(level < 0) level = guistack.length();
+    loopvrev(guistack)
+    {
+       menu *m = guistack[i];
+       if(m->onclear)
+       {
+           uint *action = m->onclear;
+           m->onclear = NULL;
+           execute(action);
+           freecode(action);
+       }
+    }
+    cleargui(level);
+}
+
 void guionclear(char *action)
 {
     if(guistack.empty()) return;
     menu *m = guistack.last();
-    DELETEA(m->onclear);
-    if(action[0]) m->onclear = newstring(action);
+    if(m->onclear) { freecode(m->onclear); m->onclear = NULL; } 
+    if(action[0]) m->onclear = compilecode(action);
 }
 
-void guistayopen(char *contents)
+void guistayopen(uint *contents)
 {
     bool oldclearmenu = shouldclearmenu;
     shouldclearmenu = false;
@@ -129,7 +225,7 @@ void guistayopen(char *contents)
     shouldclearmenu = oldclearmenu;
 }
 
-void guinoautotab(char *contents)
+void guinoautotab(uint *contents)
 {
     if(!cgui) return;
     cgui->allowautotab(false);
@@ -141,10 +237,11 @@ void guinoautotab(char *contents)
 void guibutton(char *name, char *action, char *icon)
 {
     if(!cgui) return;
-    int ret = cgui->button(name, GUI_BUTTON_COLOR, *icon ? icon : (strstr(action, "showgui") ? "menu" : "action"));
+    bool hideicon = !strcmp(icon, "0");
+    int ret = cgui->button(name, GUI_BUTTON_COLOR, hideicon ? NULL : (icon[0] ? icon : (strstr(action, "showgui") ? "menu" : "action")));
     if(ret&G3D_UP) 
     {
-        executelater.add(newstring(*action ? action : name));
+        updatelater.add().schedule(action[0] ? action : name);
         if(shouldclearmenu) clearlater = true;
     }
     else if(ret&G3D_ROLLOVER)
@@ -168,7 +265,7 @@ void guiimage(char *path, char *action, float *scale, int *overlaid, char *alt)
     {
         if(*action)
         {
-            executelater.add(newstring(action));
+            updatelater.add().schedule(action);
             if(shouldclearmenu) clearlater = true;
         }
     }
@@ -195,7 +292,8 @@ void guitextbox(char *text, int *width, int *height, int *color)
 
 void guitext(char *name, char *icon)
 {
-    if(cgui) cgui->text(name, icon[0] ? GUI_BUTTON_COLOR : GUI_TEXT_COLOR, icon[0] ? icon : "info");
+    bool hideicon = !strcmp(icon, "0");
+    if(cgui) cgui->text(name, !hideicon && icon[0] ? GUI_BUTTON_COLOR : GUI_TEXT_COLOR, hideicon ? NULL : (icon[0] ? icon : "info"));
 }
 
 void guititle(char *name)
@@ -213,46 +311,77 @@ void guibar()
     if(cgui) cgui->separator();
 }
 
-static void updateval(char *var, int val, char *onchange)
+void guistrut(float *strut, int *alt)
 {
-    ident *id = getident(var);
-    string assign;
-    if(!id) return;
-    switch(id->type)
+    if(cgui)
     {
-        case ID_VAR:
-        case ID_FVAR:
-        case ID_SVAR:
-            formatstring(assign)("%s %d", var, val);
-            break;
-        case ID_ALIAS: 
-            formatstring(assign)("%s = %d", var, val);
-            break;
-        default:
-            return;
+        if(*alt) cgui->strut(*strut); else cgui->space(*strut);
     }
-    executelater.add(newstring(assign));
-    if(onchange[0]) executelater.add(newstring(onchange)); 
+}
+
+void guispring(int *weight)
+{
+    if(cgui) cgui->spring(max(*weight, 1));
+}
+
+void guicolumn(int *col)
+{
+    if(cgui) cgui->column(*col);
+}
+
+template<class T> static void updateval(char *var, T val, char *onchange)
+{
+    ident *id = writeident(var);
+    updatelater.add().schedule(id, val);
+    if(onchange[0]) updatelater.add().schedule(onchange);
 }
 
 static int getval(char *var)
 {
-    ident *id = getident(var);
+    ident *id = readident(var);
     if(!id) return 0;
     switch(id->type)
     {
         case ID_VAR: return *id->storage.i;
         case ID_FVAR: return int(*id->storage.f);
-        case ID_SVAR: return atoi(*id->storage.s);
-        case ID_ALIAS: return atoi(id->action);
+        case ID_SVAR: return parseint(*id->storage.s);
+        case ID_ALIAS: return id->getint();
         default: return 0;
     }
 }
 
+static float getfval(char *var)
+{
+    ident *id = readident(var);
+    if(!id) return 0;
+    switch(id->type)
+    {
+        case ID_VAR: return *id->storage.i;
+        case ID_FVAR: return *id->storage.f;
+        case ID_SVAR: return parsefloat(*id->storage.s);
+        case ID_ALIAS: return id->getfloat();
+        default: return 0;
+    }
+}
+
+static const char *getsval(char *var)
+{
+    ident *id = readident(var);
+    if(!id) return "";
+    switch(id->type)
+    {
+        case ID_VAR: return intstr(*id->storage.i);
+        case ID_FVAR: return floatstr(*id->storage.f);
+        case ID_SVAR: return *id->storage.s;
+        case ID_ALIAS: return id->getstr();
+        default: return "";
+    }
+}
+
 void guislider(char *var, int *min, int *max, char *onchange)
 {
 	if(!cgui) return;
-    int oldval = getval(var), val = oldval, vmin = *max ? *min : getvarmin(var), vmax = *max ? *max : getvarmax(var);
+    int oldval = getval(var), val = oldval, vmin = *max > INT_MIN ? *min : getvarmin(var), vmax = *max > INT_MIN ? *max : getvarmax(var);
     cgui->slider(val, vmin, vmax, GUI_TITLE_COLOR);
     if(val != oldval) updateval(var, val, onchange);
 }
@@ -264,30 +393,49 @@ void guilistslider(char *var, char *list, char *onchange)
     list += strspn(list, "\n\t ");
     while(*list)
     {
-        vals.add(atoi(list));
+        vals.add(parseint(list));
         list += strcspn(list, "\n\t \0");
         list += strspn(list, "\n\t ");
     }
     if(vals.empty()) return;
     int val = getval(var), oldoffset = vals.length()-1, offset = oldoffset;
     loopv(vals) if(val <= vals[i]) { oldoffset = offset = i; break; }
-    defformatstring(label)("%d", val);
-    cgui->slider(offset, 0, vals.length()-1, GUI_TITLE_COLOR, label);
+    cgui->slider(offset, 0, vals.length()-1, GUI_TITLE_COLOR, intstr(val));
     if(offset != oldoffset) updateval(var, vals[offset], onchange);
 }
 
-void guicheckbox(char *name, char *var, int *on, int *off, char *onchange)
+void guinameslider(char *var, char *names, char *list, char *onchange)
 {
-    bool enabled = getval(var)!=*off;
+    if(!cgui) return;
+    vector<int> vals;
+    list += strspn(list, "\n\t ");
+    while(*list)
+    {
+        vals.add(parseint(list));
+        list += strcspn(list, "\n\t \0");
+        list += strspn(list, "\n\t ");
+    }
+    if(vals.empty()) return;
+    int val = getval(var), oldoffset = vals.length()-1, offset = oldoffset;
+    loopv(vals) if(val <= vals[i]) { oldoffset = offset = i; break; }
+    char *label = indexlist(names, offset);
+    cgui->slider(offset, 0, vals.length()-1, GUI_TITLE_COLOR, label);
+    if(offset != oldoffset) updateval(var, vals[offset], onchange);
+    delete[] label;
+}
+
+void guicheckbox(char *name, char *var, float *on, float *off, char *onchange)
+{
+    bool enabled = getfval(var)!=*off;
     if(cgui && cgui->button(name, GUI_BUTTON_COLOR, enabled ? "checkbox_on" : "checkbox_off")&G3D_UP)
     {
-        updateval(var, enabled ? *off : (*on || *off ? *on : 1), onchange);
+        updateval(var, enabled ? *off : (*on || *off ? *on : 1.0f), onchange);
     }
 }
 
-void guiradio(char *name, char *var, int *n, char *onchange)
+void guiradio(char *name, char *var, float *n, char *onchange)
 {
-    bool enabled = getval(var)==*n;
+    bool enabled = getfval(var)==*n;
     if(cgui && cgui->button(name, GUI_BUTTON_COLOR, enabled ? "radio_on" : "radio_off")&G3D_UP)
     {
         if(!enabled) updateval(var, *n, onchange);
@@ -308,15 +456,9 @@ void guibitfield(char *name, char *var, int *mask, char *onchange)
 void guifield(char *var, int *maxlength, char *onchange)
 {   
     if(!cgui) return;
-    const char *initval = "";
-    ident *id = getident(var);
-    if(id && id->type==ID_ALIAS) initval = id->action;
+    const char *initval = getsval(var);
 	char *result = cgui->field(var, GUI_BUTTON_COLOR, *maxlength ? *maxlength : 12, 0, initval);
-    if(result) 
-    {
-        alias(var, result);
-        if(onchange[0]) executelater.add(newstring(onchange));
-    }
+    if(result) updateval(var, result, onchange); 
 }
 
 //-ve maxlength indicates a wrapped text field of any (approx 260 chars) length, |maxlength| is the field width
@@ -331,21 +473,15 @@ void guieditor(char *name, int *maxlength, int *height, int *mode)
 void guikeyfield(char *var, int *maxlength, char *onchange)
 {
     if(!cgui) return;
-    const char *initval = "";
-    ident *id = getident(var);
-    if(id && id->type==ID_ALIAS) initval = id->action;
+    const char *initval = getsval(var);
     char *result = cgui->keyfield(var, GUI_BUTTON_COLOR, *maxlength ? *maxlength : -8, 0, initval);
-    if(result)
-    {
-        alias(var, result);
-        if(onchange[0]) execute(onchange);
-    }
+    if(result) updateval(var, result, onchange);
 }
 
 //use text<action> to do more...
 
 
-void guilist(char *contents)
+void guilist(uint *contents)
 {
     if(!cgui) return;
     cgui->pushlist();
@@ -353,7 +489,17 @@ void guilist(char *contents)
     cgui->poplist();
 }
 
-void newgui(char *name, char *contents, char *header)
+void guialign(int *align, uint *contents)
+{
+    if(!cgui) return;
+    cgui->pushlist();
+    if(*align >= 0) cgui->spring();
+    execute(contents);
+    if(*align == 0) cgui->spring(); 
+    cgui->poplist();
+}
+
+void newgui(char *name, char *contents, char *header, char *init)
 {
     menu *m = guis.access(name);
     if(!m)
@@ -365,45 +511,84 @@ void newgui(char *name, char *contents, char *header)
     else
     {
         DELETEA(m->header);
-        DELETEA(m->contents);
+        freecode(m->contents);
+        freecode(m->init);
     }
-    m->header = header && header[0] ? newstring(header) : NULL;
-    m->contents = newstring(contents);
+    if(header && header[0])
+    {
+        char *end = NULL;
+        int val = strtol(header, &end, 0);
+        if(end && !*end)
+        {
+            m->header = NULL;
+            m->showtab = val != 0;
+        }
+        else
+        {
+            m->header = newstring(header);
+            m->showtab = true;
+        }
+    }
+    else
+    {
+        m->header = NULL;
+        m->showtab = true;
+    }
+    m->contents = compilecode(contents);
+    m->init = init && init[0] ? compilecode(init) : NULL;
 }
 
-void guiservers()
+menu *guiserversmenu = NULL;
+
+void guiservers(uint *header, int *pagemin, int *pagemax)
 {
-    extern char *showservers(g3d_gui *cgui);
+    extern const char *showservers(g3d_gui *cgui, uint *header, int pagemin, int pagemax);
     if(cgui) 
     {
-        char *command = showservers(cgui);
+        const char *command = showservers(cgui, header, *pagemin, *pagemax > 0 ? *pagemax : INT_MAX);
         if(command)
         {
-            executelater.add(command);
+            updatelater.add().schedule(command);
             if(shouldclearmenu) clearlater = true;
+            guiserversmenu = clearlater || guistack.empty() ? NULL : guistack.last();
         }
     }
 }
 
-COMMAND(newgui, "sss");
+void notifywelcome()
+{
+    if(guiserversmenu)
+    {
+        if(guistack.length() && guistack.last() == guiserversmenu) clearguis();
+        guiserversmenu = NULL;
+    }
+}
+ 
+COMMAND(newgui, "ssss");
 COMMAND(guibutton, "sss");
 COMMAND(guitext, "ss");
-COMMAND(guiservers, "s");
+COMMAND(guiservers, "eii");
 ICOMMAND(cleargui, "i", (int *n), intret(cleargui(*n)));
 COMMAND(showgui, "s");
+COMMAND(hidegui, "s");
 COMMAND(guionclear, "s");
-COMMAND(guistayopen, "s");
-COMMAND(guinoautotab, "s");
+COMMAND(guistayopen, "e");
+COMMAND(guinoautotab, "e");
 
-COMMAND(guilist, "s");
+COMMAND(guilist, "e");
+COMMAND(guialign, "ie");
 COMMAND(guititle, "s");
 COMMAND(guibar,"");
+COMMAND(guistrut,"fi");
+COMMAND(guispring, "i");
+COMMAND(guicolumn, "i");
 COMMAND(guiimage,"ssfis");
-COMMAND(guislider,"siis");
+COMMAND(guislider,"sbbs");
 COMMAND(guilistslider, "sss");
-COMMAND(guiradio,"ssis");
+COMMAND(guinameslider, "ssss");
+COMMAND(guiradio,"ssfs");
 COMMAND(guibitfield, "ssis");
-COMMAND(guicheckbox, "ssiis");
+COMMAND(guicheckbox, "ssffs");
 COMMAND(guitab, "s");
 COMMAND(guifield, "sis");
 COMMAND(guikeyfield, "sis");
@@ -411,6 +596,52 @@ COMMAND(guieditor, "siii");
 COMMAND(guicolor, "i");
 COMMAND(guitextbox, "siii");
 
+void guiplayerpreview(int *model, int *team, int *weap, char *action, float *scale, int *overlaid)
+{
+    if(!cgui) return;
+    int ret = cgui->playerpreview(*model, *team, *weap, *scale, *overlaid!=0);
+    if(ret&G3D_UP)
+    {
+        if(*action)
+        {
+            updatelater.add().schedule(action);
+            if(shouldclearmenu) clearlater = true;
+        }
+    }
+}
+COMMAND(guiplayerpreview, "iiisfi");
+
+void guimodelpreview(char *model, char *animspec, char *action, float *scale, int *overlaid)
+{
+    if(!cgui) return;
+    int anim = ANIM_ALL;
+    if(animspec[0])
+    {
+        if(isdigit(animspec[0])) 
+        {
+            anim = parseint(animspec);
+            if(anim >= 0) anim %= ANIM_INDEX;
+            else anim = ANIM_ALL;
+        }
+        else
+        {
+            vector<int> anims;
+            findanims(animspec, anims);
+            if(anims.length()) anim = anims[0];
+        }
+    }
+    int ret = cgui->modelpreview(model, anim|ANIM_LOOP, *scale, *overlaid!=0);
+    if(ret&G3D_UP)
+    {
+        if(*action)
+        {
+            updatelater.add().schedule(action);
+            if(shouldclearmenu) clearlater = true;
+        }
+    }
+}
+COMMAND(guimodelpreview, "sssfi");
+
 struct change
 {
     int type;
@@ -435,8 +666,8 @@ static struct applymenu : menu
         {
             int changetypes = 0;
             loopv(needsapply) changetypes |= needsapply[i].type;
-            if(changetypes&CHANGE_GFX) executelater.add(newstring("resetgl"));
-            if(changetypes&CHANGE_SOUND) executelater.add(newstring("resetsound"));
+            if(changetypes&CHANGE_GFX) updatelater.add().schedule("resetgl");
+            if(changetypes&CHANGE_SOUND) updatelater.add().schedule("resetsound");
             clearlater = true;
         }
         if(g.button("no", GUI_BUTTON_COLOR, "action")&G3D_UP)
@@ -447,7 +678,7 @@ static struct applymenu : menu
     void clear()
     {
         menu::clear();
-        needsapply.setsize(0);
+        needsapply.shrink(0);
     }
 } applymenu;
 
@@ -481,25 +712,11 @@ void menuprocess()
 {
     processingmenu = true;
     int wasmain = mainmenu, level = guistack.length();
-    loopv(executelater) execute(executelater[i]);
-    executelater.deletecontentsa();
+    loopv(updatelater) updatelater[i].run();
+    updatelater.shrink(0);
     if(wasmain > mainmenu || clearlater)
     {
-        if(wasmain > mainmenu || level==guistack.length()) 
-        {
-            loopvrev(guistack)
-            {
-                menu *m = guistack[i];
-                if(m->onclear) 
-                {
-                    char *action = m->onclear;
-                    m->onclear = NULL;
-                    execute(action);
-                    delete[] action;
-                }
-            }
-            cleargui(level); 
-        }
+        if(wasmain > mainmenu || level==guistack.length()) clearguis(level); 
         clearlater = false;
     }
     if(mainmenu && !isconnected(true) && guistack.empty()) showgui("main");
@@ -510,7 +727,7 @@ VAR(mainmenu, 1, 1, 0);
 
 void clearmainmenu()
 {
-    if(mainmenu && (isconnected() || haslocalclients()))
+    if(mainmenu && isconnected())
     {
         mainmenu = 0;
         if(!processingmenu) cleargui();
