@@ -9,10 +9,35 @@ model *loadingmodel = NULL;
 #include "animmodel.h"
 #include "vertmodel.h"
 #include "skelmodel.h"
+
+static model *(__cdecl *modeltypes[NUMMODELTYPES])(const char *);
+
+static int addmodeltype(int type, model *(__cdecl *loader)(const char *))
+{
+    modeltypes[type] = loader;
+    return type;
+}
+
+#define MODELTYPE(modeltype, modelclass) \
+static model *__loadmodel__##modelclass(const char *filename) \
+{ \
+    return new modelclass(filename); \
+} \
+static int __dummy__##modelclass = addmodeltype((modeltype), __loadmodel__##modelclass);
+ 
 #include "md2.h"
 #include "md3.h"
 #include "md5.h"
 #include "obj.h"
+#include "smd.h"
+#include "iqm.h"
+
+MODELTYPE(MDL_MD2, md2);
+MODELTYPE(MDL_MD3, md3);
+MODELTYPE(MDL_MD5, md5);
+MODELTYPE(MDL_OBJ, obj);
+MODELTYPE(MDL_SMD, smd);
+MODELTYPE(MDL_IQM, iqm);
 
 #define checkmdl if(!loadingmodel) { conoutf(CON_ERROR, "not loading a model"); return; }
 
@@ -86,16 +111,25 @@ void mdlalphadepth(int *depth)
 
 COMMAND(mdlalphadepth, "i");
 
-void mdlglow(int *percent)
+void mdldepthoffset(int *offset)
 {
     checkmdl;
-    float glow = 3.0f;
-    if(*percent>0) glow = *percent/100.0f;
-    else if(*percent<0) glow = 0.0f;
-    loadingmodel->setglow(glow);
+    loadingmodel->depthoffset = *offset!=0;
 }
 
-COMMAND(mdlglow, "i");
+COMMAND(mdldepthoffset, "i");
+
+void mdlglow(int *percent, int *delta, float *pulse)
+{
+    checkmdl;
+    float glow = 3.0f, glowdelta = *delta/100.0f, glowpulse = *pulse > 0 ? *pulse/1000.0f : 0;
+    if(*percent>0) glow = *percent/100.0f;
+    else if(*percent<0) glow = 0.0f;
+    glowdelta -= glow;
+    loadingmodel->setglow(glow, glowdelta, glowpulse);
+}
+
+COMMAND(mdlglow, "iif");
 
 void mdlglare(float *specglare, float *glowglare)
 {
@@ -208,8 +242,8 @@ void mdlname()
 COMMAND(mdlname, "");
 
 #define checkragdoll \
-    skelmodel *m = dynamic_cast<skelmodel *>(loadingmodel); \
-    if(!m) { conoutf(CON_ERROR, "not loading a skeletal model"); return; } \
+    if(!loadingmodel->skeletal()) { conoutf(CON_ERROR, "not loading a skeletal model"); return; } \
+    skelmodel *m = (skelmodel *)loadingmodel; \
     skelmodel::skelmeshgroup *meshes = (skelmodel::skelmeshgroup *)m->parts.last()->meshes; \
     if(!meshes) return; \
     skelmodel::skeleton *skel = meshes->skel; \
@@ -218,13 +252,14 @@ COMMAND(mdlname, "");
     if(ragdoll->loaded) return;
     
 
-void rdvert(float *x, float *y, float *z)
+void rdvert(float *x, float *y, float *z, float *radius)
 {
     checkragdoll;
     ragdollskel::vert &v = ragdoll->verts.add();
     v.pos = vec(*x, *y, *z);
+    v.radius = *radius > 0 ? *radius : 1;
 }
-COMMAND(rdvert, "fff");
+COMMAND(rdvert, "ffff");
 
 void rdeye(int *v)
 {
@@ -243,17 +278,18 @@ void rdtri(int *v1, int *v2, int *v3)
 }
 COMMAND(rdtri, "iii");
 
-void rdjoint(int *n, int *t, char *v1, char *v2, char *v3)
+void rdjoint(int *n, int *t, int *v1, int *v2, int *v3)
 {
     checkragdoll;
+    if(*n < 0 || *n >= skel->numbones) return;
     ragdollskel::joint &j = ragdoll->joints.add();
     j.bone = *n;
     j.tri = *t;
-    j.vert[0] = v1[0] ? atoi(v1) : -1;
-    j.vert[1] = v2[0] ? atoi(v2) : -1;
-    j.vert[2] = v3[0] ? atoi(v3) : -1;
+    j.vert[0] = *v1;
+    j.vert[1] = *v2;
+    j.vert[2] = *v3;
 }
-COMMAND(rdjoint, "iisss");
+COMMAND(rdjoint, "iibbb");
    
 void rdlimitdist(int *v1, int *v2, float *mindist, float *maxdist)
 {
@@ -276,7 +312,14 @@ void rdlimitrot(int *t1, int *t2, float *maxangle, float *qx, float *qy, float *
     r.middle = matrix3x3(quat(*qx, *qy, *qz, *qw));
 }
 COMMAND(rdlimitrot, "iifffff");
- 
+
+void rdanimjoints(int *on)
+{
+    checkragdoll;
+    ragdoll->animjoints = *on!=0;
+}
+COMMAND(rdanimjoints, "i");
+
 // mapmodels
 
 vector<mapmodelinfo> mapmodels;
@@ -293,18 +336,18 @@ void mapmodelcompat(int *rad, int *h, int *tex, char *name, char *shadow)
     mmodel(name);
 }
 
-void mapmodelreset() 
+void mapmodelreset(int *n) 
 { 
-    if(!overrideidents && !game::allowedittoggle()) return;
-    mapmodels.setsize(0); 
+    if(!(identflags&IDF_OVERRIDDEN) && !game::allowedittoggle()) return;
+    mapmodels.shrink(clamp(*n, 0, mapmodels.length())); 
 }
 
-mapmodelinfo &getmminfo(int i) { return mapmodels.inrange(i) ? mapmodels[i] : *(mapmodelinfo *)0; }
+mapmodelinfo *getmminfo(int i) { return mapmodels.inrange(i) ? &mapmodels[i] : 0; }
 const char *mapmodelname(int i) { return mapmodels.inrange(i) ? mapmodels[i].name : NULL; }
 
 COMMAND(mmodel, "s");
 COMMANDN(mapmodel, mapmodelcompat, "iiiss");
-COMMAND(mapmodelreset, "");
+COMMAND(mapmodelreset, "i");
 ICOMMAND(mapmodelname, "i", (int *index), { result(mapmodels.inrange(*index) ? mapmodels[*index].name : ""); });
 ICOMMAND(nummapmodels, "", (), { intret(mapmodels.length()); });
 
@@ -315,18 +358,49 @@ vector<const char *> preloadmodels;
 
 void preloadmodel(const char *name)
 {
-    if(mdllookup.access(name)) return;
+    if(!name || !name[0] || mdllookup.access(name)) return;
     preloadmodels.add(newstring(name));
 }
 
-void flushpreloadedmodels()
+void flushpreloadedmodels(bool msg)
 {
     loopv(preloadmodels)
     {
         loadprogress = float(i+1)/preloadmodels.length();
-        loadmodel(preloadmodels[i], -1, true);
+        model *m = loadmodel(preloadmodels[i], -1, msg);
+        if(!m) { if(msg) conoutf(CON_WARN, "could not load model: %s", preloadmodels[i]); }
+        else
+        {
+            m->preloadmeshes();
+        }
+    }
+    preloadmodels.deletearrays();
+    loadprogress = 0;
+}
+
+void preloadusedmapmodels(bool msg, bool bih)
+{
+    vector<extentity *> &ents = entities::getents();
+    vector<int> mapmodels;
+    loopv(ents)
+    {
+        extentity &e = *ents[i];
+        if(e.type==ET_MAPMODEL && e.attr2 >= 0 && mapmodels.find(e.attr2) < 0) mapmodels.add(e.attr2);
+    }
+
+    loopv(mapmodels)
+    {
+        loadprogress = float(i+1)/mapmodels.length();
+        int mmindex = mapmodels[i];
+        mapmodelinfo *mmi = getmminfo(mmindex);
+        if(!mmi) { if(msg) conoutf(CON_WARN, "could not find map model: %d", mmindex); }
+        else if(mmi->name[0] && !loadmodel(NULL, mmindex, msg)) { if(msg) conoutf(CON_WARN, "could not load model: %s", mmi->name); }
+        else if(mmi->m)
+        {
+            if(bih) mmi->m->preloadBIH();
+            mmi->m->preloadmeshes();
+        }
     }
-    preloadmodels.deletecontentsa();
     loadprogress = 0;
 }
 
@@ -344,39 +418,24 @@ model *loadmodel(const char *name, int i, bool msg)
     if(mm) m = *mm;
     else
     { 
+        if(!name[0] || loadingmodel || lightmapping > 1) return NULL;
         if(msg)
         {
             defformatstring(filename)("packages/models/%s", name);
             renderprogress(loadprogress, filename);
         }
-        m = new md2(name);
-        loadingmodel = m;
-        if(!m->load())
+        loopi(NUMMODELTYPES)
         {
-            delete m;
-            m = new md3(name);
+            m = modeltypes[i](name);
+            if(!m) continue;
             loadingmodel = m;
-            if(!m->load())
-            {    
-                delete m;
-                m = new md5(name);
-                loadingmodel = m;
-                if(!m->load())
-                {
-                    delete m;
-                    m = new obj(name);
-                    loadingmodel = m;
-                    if(!m->load())
-                    {
-                        delete m;
-                        loadingmodel = NULL;
-                        return NULL; 
-                    }
-                }
-            }
+            if(m->load()) break;
+            DELETEP(m);
         }
         loadingmodel = NULL;
+        if(!m) return NULL;
         mdllookup.access(m->name(), m);
+        m->preloadshaders();
     }
     if(mapmodels.inrange(i) && !mapmodels[i].m) mapmodels[i].m = m;
     return m;
@@ -437,7 +496,7 @@ void render3dbox(vec &o, float tofloor, float toceil, float xradius, float yradi
     c.sub(vec(xradius, yradius, tofloor));
     float xsz = xradius*2, ysz = yradius*2;
     float h = tofloor+toceil;
-    notextureshader->set();
+    lineshader->set();
     glDisable(GL_TEXTURE_2D);
     glColor3f(1, 1, 1);
     render2dbox(c, xsz, 0, h);
@@ -451,13 +510,14 @@ void render3dbox(vec &o, float tofloor, float toceil, float xradius, float yradi
 
 void renderellipse(vec &o, float xradius, float yradius, float yaw)
 {
-    notextureshader->set();
+    lineshader->set();
     glDisable(GL_TEXTURE_2D);
     glColor3f(0.5f, 0.5f, 0.5f);
     glBegin(GL_LINE_LOOP);
-    loopi(16)
+    loopi(15)
     {
-        vec p(xradius*cosf(2*M_PI*i/16.0f), yradius*sinf(2*M_PI*i/16.0f), 0);
+        const vec2 &sc = sincos360[i*(360/15)];
+        vec p(xradius*sc.x, yradius*sc.y, 0);
         p.rotate_around_z((yaw+90)*RAD);
         p.add(o);
         glVertex3fv(p.v);
@@ -490,7 +550,7 @@ static occludequery *modelquery = NULL;
 void startmodelbatches()
 {
     numbatches = 0;
-    modelattached.setsizenodelete(0);
+    modelattached.setsize(0);
 }
 
 modelbatch &addbatchedmodel(model *m)
@@ -502,7 +562,7 @@ modelbatch &addbatchedmodel(model *m)
         if(numbatches<batches.length())
         {
             b = batches[numbatches];
-            b->batched.setsizenodelete(0);
+            b->batched.setsize(0);
         }
         else b = batches.add(new modelbatch);
         b->m = m;
@@ -518,7 +578,7 @@ void renderbatchedmodel(model *m, batchedmodel &b)
     if(b.attached>=0) a = &modelattached[b.attached];
 
     int anim = b.anim;
-    if(shadowmapping) 
+    if(shadowmapping)
     {
         anim |= ANIM_NOSKIN; 
         if(renderpath!=R_FIXEDFUNCTION) setenvparamf("shadowintensity", SHPARAM_VERTEX, 1, b.transparent);
@@ -526,6 +586,7 @@ void renderbatchedmodel(model *m, batchedmodel &b)
     else 
     {
         if(b.flags&MDL_FULLBRIGHT) anim |= ANIM_FULLBRIGHT;
+        if(b.flags&MDL_GHOST) anim |= ANIM_GHOST;
     }
 
     m->render(anim, b.basetime, b.basetime2, b.pos, b.yaw, b.pitch, b.d, a, b.color, b.dir, b.transparent);
@@ -538,11 +599,9 @@ struct transparentmodel
     float dist;
 };
 
-static int sorttransparentmodels(const transparentmodel *x, const transparentmodel *y)
+static inline bool sorttransparentmodels(const transparentmodel &x, const transparentmodel &y)
 {
-    if(x->dist > y->dist) return -1;
-    if(x->dist < y->dist) return 1;
-    return 0;
+    return x.dist < y.dist;
 }
 
 void endmodelbatches()
@@ -566,10 +625,25 @@ void endmodelbatches()
         }
         bool rendered = false;
         occludequery *query = NULL;
+        if(b.flags&MDL_GHOST)
+        {
+            loopvj(b.batched)
+            {
+                batchedmodel &bm = b.batched[j];
+                if((bm.flags&(MDL_CULL_VFC|MDL_GHOST))!=MDL_GHOST || bm.query) continue;
+                if(!rendered) { b.m->startrender(); rendered = true; }
+                renderbatchedmodel(b.m, bm);
+            }
+            if(rendered) 
+            {
+                b.m->endrender();
+                rendered = false;
+            }
+        }
         loopvj(b.batched) 
         {
             batchedmodel &bm = b.batched[j];
-            if(bm.flags&MDL_CULL_VFC) continue;
+            if(bm.flags&(MDL_CULL_VFC|MDL_GHOST)) continue;
             if(bm.query!=query)
             {
                 if(query) endquery(query);
@@ -655,7 +729,7 @@ void endmodelquery()
     }
     endquery(modelquery);
     modelquery = NULL;
-    modelattached.setsizenodelete(minattached);
+    modelattached.setsize(minattached);
 }
 
 VARP(maxmodelradiusdistance, 10, 200, 1000);
@@ -689,7 +763,7 @@ void rendermodel(entitylight *light, const char *mdl, int anim, const vec &o, fl
     if(shadowmapping && !(flags&(MDL_SHADOW|MDL_DYNSHADOW))) return;
     model *m = loadmodel(mdl); 
     if(!m) return;
-    vec center, bbradius;
+    vec center(0, 0, 0), bbradius(0, 0, 0);
     float radius = 0;
     bool shadow = !shadowmap && !glaring && (flags&(MDL_SHADOW|MDL_DYNSHADOW)) && showblobs,
          doOQ = flags&MDL_CULL_QUERY && hasOQ && oqfrags && oqdynent;
@@ -704,7 +778,7 @@ void rendermodel(entitylight *light, const char *mdl, int anim, const vec &o, fl
         }
         else
         {
-            center.rotate_around_z((-180-yaw)*RAD);
+            center.rotate_around_z(yaw*RAD);
             center.add(o);
         }
         if(flags&MDL_CULL_DIST && center.dist(camera1->o)/radius>maxmodelradiusdistance) return;
@@ -718,12 +792,12 @@ void rendermodel(entitylight *light, const char *mdl, int anim, const vec &o, fl
                 }
                 else
                 {
-                    if(fogging && center.z+radius<reflectz-waterfog) return;
+                    if(fogging && center.z+radius<reflectz-refractfog) return;
                     if(!shadow && center.z-radius>=reflectz) return;
                 }
                 if(center.dist(camera1->o)-radius>reflectdist) return;
             }
-            if(isvisiblesphere(radius, center) >= VFC_FOGGED) return;
+            if(isfoggedsphere(radius, center)) return;
             if(shadowmapping && !isshadowmapcaster(center, radius)) return;
         }
         if(shadowmapping)
@@ -756,7 +830,7 @@ void rendermodel(entitylight *light, const char *mdl, int anim, const vec &o, fl
     }
 
     if(flags&MDL_NORENDER) anim |= ANIM_NORENDER;
-    else if(showboundingbox && !shadowmapping && !reflecting && !refracting)
+    else if(showboundingbox && !shadowmapping && !reflecting && !refracting && editmode)
     {
         if(d && showboundingbox==1) 
         {
@@ -789,8 +863,9 @@ void rendermodel(entitylight *light, const char *mdl, int anim, const vec &o, fl
                     pos = d->ragdoll->center;
                     pos.z += radius/2;
                 }
-                lightreaching(pos, light->color, light->dir);
-                dynlightreaching(pos, light->color, light->dir);
+                else if(d->type < ENT_CAMERA) pos.z += 0.75f*(d->eyeheight + d->aboveeye);
+                lightreaching(pos, light->color, light->dir, (flags&MDL_LIGHT_FAST)!=0);
+                dynlightreaching(pos, light->color, light->dir, (flags&MDL_HUD)!=0);
                 game::lighteffects(d, light->color, light->dir);
                 light->millis = lastmillis;
             }
@@ -799,18 +874,18 @@ void rendermodel(entitylight *light, const char *mdl, int anim, const vec &o, fl
         {
             if(!light) 
             {
-                lightreaching(pos, lightcolor, lightdir);
-                dynlightreaching(pos, lightcolor, lightdir);
+                lightreaching(pos, lightcolor, lightdir, (flags&MDL_LIGHT_FAST)!=0);
+                dynlightreaching(pos, lightcolor, lightdir, (flags&MDL_HUD)!=0);
             }
             else if(light->millis!=lastmillis)
             {
-                lightreaching(pos, light->color, light->dir);
-                dynlightreaching(pos, light->color, light->dir);
+                lightreaching(pos, light->color, light->dir, (flags&MDL_LIGHT_FAST)!=0);
+                dynlightreaching(pos, light->color, light->dir, (flags&MDL_HUD)!=0);
                 light->millis = lastmillis;
             }
         }
         if(light) { lightcolor = light->color; lightdir = light->dir; }
-        if(flags&MDL_DYNLIGHT) dynlightreaching(pos, lightcolor, lightdir);
+        if(flags&MDL_DYNLIGHT) dynlightreaching(pos, lightcolor, lightdir, (flags&MDL_HUD)!=0);
     }
 
     if(a) for(int i = 0; a[i].tag; i++)
@@ -866,6 +941,7 @@ void rendermodel(entitylight *light, const char *mdl, int anim, const vec &o, fl
     else 
     {
         if(flags&MDL_FULLBRIGHT) anim |= ANIM_FULLBRIGHT;
+        if(flags&MDL_GHOST) anim |= ANIM_GHOST;
     }
 
     if(doOQ)
@@ -900,8 +976,8 @@ bool matchanim(const char *name, const char *pattern)
             if(!c || c=='|') break;
             else if(c=='*') 
             {
-                if(!*s || isspace(*s)) break;
-                do s++; while(*s && !isspace(*s));
+                if(!*s || iscubespace(*s)) break;
+                do s++; while(*s && !iscubespace(*s));
             }
             else if(c!=*s) break;
             else s++;
@@ -915,33 +991,36 @@ bool matchanim(const char *name, const char *pattern)
 
 void findanims(const char *pattern, vector<int> &anims)
 {
-    static const char *names[] = 
-    { 
-        "dead", "dying", "idle", 
-        "forward", "backward", "left", "right", 
-        "hold 1", "hold 2", "hold 3", "hold 4", "hold 5", "hold 6", "hold 7",
-        "attack 1", "attack 2", "attack 3", "attack 4", "attack 5", "attack 6", "attack 7",
-        "pain", 
-        "jump", "sink", "swim", 
-        "edit", "lag", "taunt", "win", "lose", 
-        "gun idle", "gun shoot",
-        "vwep idle", "vwep shoot", "shield", "powerup", 
-        "mapmodel", "trigger"
-    };
-    loopi(sizeof(names)/sizeof(names[0])) if(matchanim(names[i], pattern)) anims.add(i);
+    loopi(sizeof(animnames)/sizeof(animnames[0])) if(matchanim(animnames[i], pattern)) anims.add(i);
 }
 
+ICOMMAND(findanims, "s", (char *name),
+{
+    vector<int> anims;
+    findanims(name, anims);
+    vector<char> buf;
+    string num;
+    loopv(anims)
+    {
+        formatstring(num)("%d", anims[i]);
+        if(i > 0) buf.add(' ');
+        buf.put(num, strlen(num));
+    }
+    buf.add('\0');
+    result(buf.getbuf());
+});
+
 void loadskin(const char *dir, const char *altdir, Texture *&skin, Texture *&masks) // model skin sharing
 {
 #define ifnoload(tex, path) if((tex = textureload(path, 0, true, false))==notexture)
-#define tryload(tex, cmd, name) \
-    ifnoload(tex, makerelpath(mdir, name ".jpg", NULL, cmd)) \
+#define tryload(tex, prefix, cmd, name) \
+    ifnoload(tex, makerelpath(mdir, name ".jpg", prefix, cmd)) \
     { \
-        ifnoload(tex, makerelpath(mdir, name ".png", NULL, cmd)) \
+        ifnoload(tex, makerelpath(mdir, name ".png", prefix, cmd)) \
         { \
-            ifnoload(tex, makerelpath(maltdir, name ".jpg", NULL, cmd)) \
+            ifnoload(tex, makerelpath(maltdir, name ".jpg", prefix, cmd)) \
             { \
-                ifnoload(tex, makerelpath(maltdir, name ".png", NULL, cmd)) return; \
+                ifnoload(tex, makerelpath(maltdir, name ".png", prefix, cmd)) return; \
             } \
         } \
     }
@@ -949,8 +1028,8 @@ void loadskin(const char *dir, const char *altdir, Texture *&skin, Texture *&mas
     defformatstring(mdir)("packages/models/%s", dir);
     defformatstring(maltdir)("packages/models/%s", altdir);
     masks = notexture;
-    tryload(skin, NULL, "skin");
-    tryload(masks, "<ffmask:25>", "masks");
+    tryload(skin, NULL, NULL, "skin");
+    tryload(masks, "<stub>", NULL, "masks");
 }
 
 // convenient function that covers the usual anims for players/monsters/npcs
@@ -969,17 +1048,17 @@ void renderclient(dynent *d, const char *mdlname, modelattach *attachments, int
     if(animoverride) anim = (animoverride<0 ? ANIM_ALL : animoverride)|ANIM_LOOP;
     else if(d->state==CS_DEAD)
     {
-        anim = ANIM_DYING;
+        anim = ANIM_DYING|ANIM_NOPITCH;
         basetime = lastpain;
         if(ragdoll)
         {
-            if(!d->ragdoll || d->ragdoll->millis < basetime) anim |= ANIM_RAGDOLL;
-        }
-        else 
-        {
-            pitch *= max(1.0f - (lastmillis-basetime)/500.0f, 0.0f);
-            if(lastmillis-basetime>1000) anim = ANIM_DEAD|ANIM_LOOP;
+            if(!d->ragdoll || d->ragdoll->millis < basetime) 
+            {
+                DELETEP(d->ragdoll);
+                anim |= ANIM_RAGDOLL;
+            }
         }
+        else if(lastmillis-basetime>1000) anim = ANIM_DEAD|ANIM_LOOP|ANIM_NOPITCH;
     }
     else if(d->state==CS_EDITING || d->state==CS_SPECTATOR) anim = ANIM_EDIT|ANIM_LOOP;
     else if(d->state==CS_LAGGED)                            anim = ANIM_LAG|ANIM_LOOP;
@@ -1007,7 +1086,7 @@ void renderclient(dynent *d, const char *mdlname, modelattach *attachments, int
         
         if((anim&ANIM_INDEX)==ANIM_IDLE && (anim>>ANIM_SECONDARY)&ANIM_INDEX) anim >>= ANIM_SECONDARY;
     }
-    if(d->ragdoll && (!ragdoll || anim!=ANIM_DYING)) DELETEP(d->ragdoll);
+    if(d->ragdoll && (!ragdoll || (anim&ANIM_INDEX)!=ANIM_DYING)) DELETEP(d->ragdoll);
     if(!((anim>>ANIM_SECONDARY)&ANIM_INDEX)) anim |= (ANIM_IDLE|ANIM_LOOP)<<ANIM_SECONDARY;
     int flags = MDL_LIGHT;
     if(d!=player && !(anim&ANIM_RAGDOLL)) flags |= MDL_CULL_VFC | MDL_CULL_OCCLUDED | MDL_CULL_QUERY;
@@ -1015,6 +1094,7 @@ void renderclient(dynent *d, const char *mdlname, modelattach *attachments, int
     else flags |= MDL_CULL_DIST;
     if(d->state==CS_LAGGED) fade = min(fade, 0.3f);
     else flags |= MDL_DYNSHADOW;
+    if(modelpreviewing) flags &= ~(MDL_LIGHT | MDL_FULLBRIGHT | MDL_CULL_VFC | MDL_CULL_OCCLUDED | MDL_CULL_QUERY | MDL_CULL_DIST | MDL_DYNSHADOW);
     rendermodel(NULL, mdlname, anim, o, yaw, pitch, flags, d, attachments, basetime, 0, fade);
 }
 
@@ -1026,13 +1106,13 @@ void setbbfrommodel(dynent *d, const char *mdl)
     m->collisionbox(0, center, radius);
     if(d->type==ENT_INANIMATE && !m->ellipsecollide)
     {
-        d->collidetype = COLLIDE_AABB;
-        rotatebb(center, radius, int(d->yaw));
+        d->collidetype = COLLIDE_OBB;
+        //d->collidetype = COLLIDE_AABB;
+        //rotatebb(center, radius, int(d->yaw));
     }
     d->xradius   = radius.x + fabs(center.x);
     d->yradius   = radius.y + fabs(center.y);
-    d->radius    = max(d->xradius, d->yradius);
-    if(d->collidetype!=COLLIDE_ELLIPSE) d->xradius = d->yradius = d->radius;
+    d->radius    = d->collidetype==COLLIDE_OBB ? sqrtf(d->xradius*d->xradius + d->yradius*d->yradius) : max(d->xradius, d->yradius);
     d->eyeheight = (center.z-radius.z) + radius.z*2*m->eyeheight;
     d->aboveeye  = radius.z*2*(1.0f-m->eyeheight);
 }
