@@ -4,6 +4,10 @@
 #ifdef SERVMODE
 struct captureservmode : servmode
 #else
+VARP(capturetether, 0, 1, 1);
+VARP(autorepammo, 0, 1, 1);
+VARP(basenumbers, 0, 0, 1);
+
 struct captureclientmode : clientmode
 #endif
 {
@@ -22,6 +26,7 @@ struct captureclientmode : clientmode
     static const int MAXAMMO = 5;
     static const int REPAMMODIST = 32;
     static const int RESPAWNSECS = 5;
+    static const int MAXBASES = 100;
 
     struct baseinfo
     {
@@ -36,6 +41,8 @@ struct captureclientmode : clientmode
 
         baseinfo() { reset(); }
 
+        bool valid() const { return ammotype>0 && ammotype<=I_CARTRIDGES-I_SHELLS+1; }
+
         void noenemy()
         {
             enemy[0] = '\0';
@@ -83,12 +90,12 @@ struct captureclientmode : clientmode
 
         bool leave(const char *team)
         {
-            if(!strcmp(owner, team))
+            if(!strcmp(owner, team) && owners > 0)
             {
                 owners--;
                 return false;
             }
-            if(strcmp(enemy, team)) return false;
+            if(strcmp(enemy, team) || enemies <= 0) return false;
             enemies--;
             return !enemies;
         }
@@ -136,8 +143,8 @@ struct captureclientmode : clientmode
 
     void resetbases()
     {
-        bases.setsize(0);
-        scores.setsize(0);
+        bases.shrink(0);
+        scores.shrink(0);
         captures = 0;
     }
 
@@ -173,9 +180,10 @@ struct captureclientmode : clientmode
 
     void addbase(int ammotype, const vec &o)
     {
+        if(bases.length() >= MAXBASES) return;
         baseinfo &b = bases.add();
         b.ammogroup = min(ammotype, 0);
-        b.ammotype = ammotype > 0 ? ammotype : rnd(5)+1;
+        b.ammotype = ammotype > 0 ? ammotype : rnd(I_GRENADES-I_SHELLS+1)+1;
         b.o = o;
 
         if(b.ammogroup)
@@ -185,7 +193,8 @@ struct captureclientmode : clientmode
                 b.ammotype = bases[i].ammotype;
                 return;
             }
-            int uses[5] = { 0, 0, 0, 0, 0 };
+            int uses[I_GRENADES-I_SHELLS+1];
+            memset(uses, 0, sizeof(uses));
             loopi(bases.length()-1) if(bases[i].ammogroup)
             {
                 loopj(i) if(bases[j].ammogroup == bases[i].ammogroup) goto nextbase;
@@ -193,9 +202,9 @@ struct captureclientmode : clientmode
                 nextbase:;
             }
             int mintype = 0;
-            loopi(5) if(uses[i] < uses[mintype]) mintype = i;
-            int numavail = 0, avail[5];
-            loopi(5) if(uses[i] == uses[mintype]) avail[numavail++] = i+1;
+            loopi(I_GRENADES-I_SHELLS+1) if(uses[i] < uses[mintype]) mintype = i;
+            int numavail = 0, avail[I_GRENADES-I_SHELLS+1];
+            loopi(I_GRENADES-I_SHELLS+1) if(uses[i] == uses[mintype]) avail[numavail++] = i+1;
             b.ammotype = avail[rnd(numavail)];
         }
     }
@@ -240,16 +249,10 @@ struct captureclientmode : clientmode
     }
 
 #ifndef SERVMODE
-    static const int FIREBALLRADIUS = 5;
+    static const int AMMOHEIGHT = 5;
 
-    float radarscale;
-
-    IVARP(capturetether, 0, 1, 1);
-    IVARP(autorepammo, 0, 1, 1);
-
-    captureclientmode() : captures(0), radarscale(0)
+    captureclientmode() : captures(0)
     {
-        CCOMMAND(repammo, "", (captureclientmode *self), self->replenishammo());
     }
 
     void respawned(fpsent *d)
@@ -262,9 +265,9 @@ struct captureclientmode : clientmode
         loopv(bases)
         {
             baseinfo &b = bases[i];
-            if(b.ammotype>0 && b.ammotype<=I_CARTRIDGES-I_SHELLS+1 && insidebase(b, player1->feetpos()) && player1->hasmaxammo(b.ammotype-1+I_SHELLS)) return;
+            if(b.valid() && insidebase(b, player1->feetpos()) && player1->hasmaxammo(b.ammotype-1+I_SHELLS)) return;
         }
-        addmsg(SV_REPAMMO, "rc", player1);
+        addmsg(N_REPAMMO, "rc", player1);
     }
 
     void receiveammo(fpsent *d, int type)
@@ -272,20 +275,22 @@ struct captureclientmode : clientmode
         type += I_SHELLS-1;
         if(type<I_SHELLS || type>I_CARTRIDGES) return;
         entities::repammo(d, type, d==player1);
+        int icon = itemstats[type-I_SHELLS].icon;
+        if(icon >= 0) particle_icon(d->abovehead(), icon%4, icon/4, PART_HUD_ICON_GREY, 2000, 0xFFFFFF, 2.0f, -8);
     }
 
     void checkitems(fpsent *d)
     {
-        if(m_regencapture || !autorepammo() || d!=player1 || d->state!=CS_ALIVE) return;
+        if(m_regencapture || !autorepammo || d!=player1 || d->state!=CS_ALIVE) return;
         vec o = d->feetpos();
         loopv(bases)
         {
             baseinfo &b = bases[i];
-            if(b.ammotype>0 && b.ammotype<=I_CARTRIDGES-I_SHELLS+1 && insidebase(b, d->feetpos()) && !strcmp(b.owner, d->team) && b.o.dist(o) < 12)
+            if(b.valid() && insidebase(b, d->feetpos()) && !strcmp(b.owner, d->team) && b.o.dist(o) < 12)
             {
                 if(d->lastrepammo!=i)
                 {
-                    if(b.ammo > 0 && !player1->hasmaxammo(b.ammotype-1+I_SHELLS)) addmsg(SV_REPAMMO, "rc", d);
+                    if(b.ammo > 0 && !player1->hasmaxammo(b.ammotype-1+I_SHELLS)) addmsg(N_REPAMMO, "rc", d);
                     d->lastrepammo = i;
                 }
                 return;
@@ -304,20 +309,21 @@ struct captureclientmode : clientmode
             loopv(bases)
             {
                 baseinfo &b = bases[i];
-                if(!insidebase(b, d->feetpos()) || (strcmp(b.owner, d->team) && strcmp(b.enemy, d->team))) continue;
-                particle_flare(b.ammopos, pos, 0, PART_LIGHTNING, strcmp(d->team, player1->team) ? 0xFF2222 : 0x2222FF, 0.28f);
+                if(!b.valid() || !insidebase(b, d->feetpos()) || (strcmp(b.owner, d->team) && strcmp(b.enemy, d->team))) continue;
+                if(d->lastbase < 0 && (lookupmaterial(d->feetpos())&MATF_CLIP) == MAT_GAMECLIP) break;
+                particle_flare(pos, vec(b.ammopos.x, b.ammopos.y, b.ammopos.z - AMMOHEIGHT - 4.4f), 0, PART_LIGHTNING, strcmp(d->team, player1->team) ? 0xFF2222 : 0x2222FF, 1.0f);
                 if(oldbase < 0)
                 {
-                    particle_fireball(pos, 4.8f, PART_EXPLOSION_NO_GLARE, 250, strcmp(d->team, player1->team) ? 0x802020 : 0x2020FF, 4.8f);
-                    particle_splash(PART_SPARK, 50, 250, pos, 0xB49B4B, 0.24f);
+                    particle_fireball(pos, 4.8f, PART_EXPLOSION, 250, strcmp(d->team, player1->team) ? 0x802020 : 0x2020FF, 4.8f);
+                    particle_splash(PART_SPARK, 50, 250, pos, strcmp(d->team, player1->team) ? 0x802020 : 0x2020FF, 0.24f);
                 }
                 d->lastbase = i;
             }
         }
         if(d->lastbase < 0 && oldbase >= 0)
         {
-            particle_fireball(pos, 4.8f, PART_EXPLOSION_NO_GLARE, 250, strcmp(d->team, player1->team) ? 0x802020 : 0x2020FF, 4.8f);
-            particle_splash(PART_SPARK, 50, 250, pos, 0xB49B4B, 0.24f);
+            particle_fireball(pos, 4.8f, PART_EXPLOSION, 250, strcmp(d->team, player1->team) ? 0x802020 : 0x2020FF, 4.8f);
+            particle_splash(PART_SPARK, 50, 250, pos, strcmp(d->team, player1->team) ? 0x802020 : 0x2020FF, 0.24f);
         }
     }
 
@@ -325,11 +331,13 @@ struct captureclientmode : clientmode
     {
         static const char *basemodels[3] = { "base/neutral", "base/red", "base/blue" };
         loopi(3) preloadmodel(basemodels[i]);
+        preloadsound(S_V_BASECAP);
+        preloadsound(S_V_BASELOST);
     }
 
     void rendergame()
     {
-        if(capturetether() && canaddparticles())
+        if(capturetether && canaddparticles())
         {
             loopv(players)
             {
@@ -341,30 +349,32 @@ struct captureclientmode : clientmode
         loopv(bases)
         {
             baseinfo &b = bases[i];
-            const char *flagname = b.owner[0] ? (strcmp(b.owner, player1->team) ? "base/red" : "base/blue") : "base/neutral";
-            rendermodel(&b.light, flagname, ANIM_MAPMODEL|ANIM_LOOP, b.o, 0, 0, MDL_SHADOW | MDL_CULL_VFC | MDL_CULL_OCCLUDED);
-            particle_fireball(b.ammopos, 4.8f, PART_EXPLOSION_NO_GLARE, 0, b.owner[0] ? (strcmp(b.owner, player1->team) ? 0x802020 : 0x2020FF) : 0x208020, 4.8f);
+            if(!b.valid()) continue;
+            const char *basename = b.owner[0] ? (strcmp(b.owner, player1->team) ? "base/red" : "base/blue") : "base/neutral";
+            rendermodel(&b.light, basename, ANIM_MAPMODEL|ANIM_LOOP, b.o, 0, 0, MDL_SHADOW | MDL_CULL_VFC | MDL_CULL_OCCLUDED);
+            float fradius = 1.0f, fheight = 0.5f;
+            regular_particle_flame(PART_FLAME, vec(b.ammopos.x, b.ammopos.y, b.ammopos.z - 4.5f), fradius, fheight, b.owner[0] ? (strcmp(b.owner, player1->team) ? 0x802020 : 0x2020FF) : 0x208020, 3, 2.0f);
+            //regular_particle_flame(PART_SMOKE, vec(b.ammopos.x, b.ammopos.y, b.ammopos.z - 4.5f + 4.0f*min(fradius, fheight)), fradius, fheight, 0x303020, 1, 4.0f, 100.0f, 2000.0f, -20);
 
-            if(b.ammotype>0 && b.ammotype<=I_CARTRIDGES-I_SHELLS+1)
+//            particle_fireball(b.ammopos, 4.8f, PART_EXPLOSION, 0, b.owner[0] ? (strcmp(b.owner, player1->team) ? 0x802020 : 0x2020FF) : 0x208020, 4.8f);
+
+            const char *ammoname = entities::entmdlname(I_SHELLS+b.ammotype-1);
+            if(m_regencapture)
             {
-                const char *ammoname = entities::entmdlname(I_SHELLS+b.ammotype-1);
-                if(m_regencapture)
-                {
-                    vec height(0, 0, 0);
-                    abovemodel(height, ammoname);
-                    vec ammopos(b.ammopos);
-                    ammopos.z -= height.z/2 + sinf(lastmillis/100.0f)/20;
-                    rendermodel(&b.light, ammoname, ANIM_MAPMODEL|ANIM_LOOP, ammopos, lastmillis/10.0f, 0, MDL_SHADOW | MDL_CULL_VFC | MDL_CULL_OCCLUDED);
-                }
-                else loopj(b.ammo)
-                {
-                    float angle = 2*M_PI*(lastmillis/4000.0f + j/float(MAXAMMO));
-                    vec ammopos(b.o);
-                    ammopos.x += 10*cosf(angle);
-                    ammopos.y += 10*sinf(angle);
-                    ammopos.z += 4;
-                    rendermodel(&b.light, entities::entmdlname(I_SHELLS+b.ammotype-1), ANIM_MAPMODEL|ANIM_LOOP, ammopos, 0, 0, MDL_SHADOW | MDL_CULL_VFC | MDL_CULL_OCCLUDED);
-                }
+                vec height(0, 0, 0);
+                abovemodel(height, ammoname);
+                vec ammopos(b.ammopos);
+                ammopos.z -= height.z/2 + sinf(lastmillis/100.0f)/20;
+                rendermodel(&b.light, ammoname, ANIM_MAPMODEL|ANIM_LOOP, ammopos, lastmillis/10.0f, 0, MDL_SHADOW | MDL_CULL_VFC | MDL_CULL_OCCLUDED);
+            }
+            else loopj(b.ammo)
+            {
+                float angle = 2*M_PI*(lastmillis/4000.0f + j/float(MAXAMMO));
+                vec ammopos(b.o);
+                ammopos.x += 10*cosf(angle);
+                ammopos.y += 10*sinf(angle);
+                ammopos.z += 4;
+                rendermodel(&b.light, entities::entmdlname(I_SHELLS+b.ammotype-1), ANIM_MAPMODEL|ANIM_LOOP, ammopos, 0, 0, MDL_SHADOW | MDL_CULL_VFC | MDL_CULL_OCCLUDED);
             }
 
             int tcolor = 0x1EC850, mtype = -1, mcolor = 0xFFFFFF, mcolor2 = 0;
@@ -372,19 +382,26 @@ struct captureclientmode : clientmode
             {
                 bool isowner = !strcmp(b.owner, player1->team);
                 if(b.enemy[0]) { mtype = PART_METER_VS; mcolor = 0xFF1932; mcolor2 = 0x3219FF; if(!isowner) swap(mcolor, mcolor2); }
-                formatstring(b.info)("%s", b.owner); tcolor = isowner ? 0x6496FF : 0xFF4B19;
+                if(!b.name[0]) formatstring(b.info)("base %d: %s", i+1, b.owner);
+                else if(basenumbers) formatstring(b.info)("%s (%d): %s", b.name, i+1, b.owner);
+                else formatstring(b.info)("%s: %s", b.name, b.owner);
+                tcolor = isowner ? 0x6496FF : 0xFF4B19;
             }
             else if(b.enemy[0])
             {
-                formatstring(b.info)("%s", b.enemy);
+                if(!b.name[0]) formatstring(b.info)("base %d: %s", i+1, b.enemy);
+                else if(basenumbers) formatstring(b.info)("%s (%d): %s", b.name, i+1, b.enemy);
+                else formatstring(b.info)("%s: %s", b.name, b.enemy);
                 if(strcmp(b.enemy, player1->team)) { tcolor = 0xFF4B19; mtype = PART_METER; mcolor = 0xFF1932; }
                 else { tcolor = 0x6496FF; mtype = PART_METER; mcolor = 0x3219FF; }
             }
-            else b.info[0] = '\0';
+            else if(!b.name[0]) formatstring(b.info)("base %d", i+1);
+            else if(basenumbers) formatstring(b.info)("%s (%d)", b.name, i+1);
+            else copystring(b.info, b.name);
 
             vec above(b.ammopos);
-            above.z += FIREBALLRADIUS+1.0f;
-            particle_text(above, b.info, PART_TEXT, 1, tcolor, 2.0f);
+            above.z += AMMOHEIGHT;
+            if(b.info[0]) particle_text(above, b.info, PART_TEXT, 1, tcolor, 2.0f);
             if(mtype>=0)
             {
                 above.z += 3.0f;
@@ -393,23 +410,14 @@ struct captureclientmode : clientmode
         }
     }
 
-    void drawradar(float x, float y, float s)
+    void drawblips(fpsent *d, float blipsize, int fw, int fh, int type, bool skipenemy = false)
     {
-        glTexCoord2f(0.0f, 0.0f); glVertex2f(x,   y);
-        glTexCoord2f(1.0f, 0.0f); glVertex2f(x+s, y);
-        glTexCoord2f(1.0f, 1.0f); glVertex2f(x+s, y+s);
-        glTexCoord2f(0.0f, 1.0f); glVertex2f(x,   y+s);
-    }
-
-    void drawblips(fpsent *d, int x, int y, int s, int type, bool skipenemy = false)
-    {
-        const char *textures[3] = {"packages/hud/blip_red.png", "packages/hud/blip_grey.png", "packages/hud/blip_blue.png"};
-        settexture(textures[max(type+1, 0)]);
-        glBegin(GL_QUADS);
-        float scale = radarscale<=0 || radarscale>maxradarscale ? maxradarscale : radarscale;
+        float scale = calcradarscale();
+        int blips = 0;
         loopv(bases)
         {
             baseinfo &b = bases[i];
+            if(!b.valid()) continue;
             if(skipenemy && b.enemy[0]) continue;
             switch(type)
             {
@@ -418,15 +426,31 @@ struct captureclientmode : clientmode
                 case -1: if(!b.owner[0] || !strcmp(b.owner, player1->team)) continue; break;
                 case -2: if(!b.enemy[0] || !strcmp(b.enemy, player1->team)) continue; break;
             }
-            vec dir(b.o);
-            dir.sub(d->o);
-            dir.z = 0.0f;
-            float dist = dir.magnitude();
-            if(dist >= scale*(1 - 0.05f)) dir.mul(scale*(1 - 0.05f)/dist);
-            dir.rotate_around_z(-d->yaw*RAD);
-            drawradar(x + s*0.5f*(1.0f + dir.x/scale - 0.05f), y + s*0.5f*(1.0f + dir.y/scale - 0.05f), 0.05f*s);
+            vec dir(d->o);
+            dir.sub(b.o).div(scale);
+            float dist = dir.magnitude2(), maxdist = 1 - 0.05f - blipsize;
+            if(dist >= maxdist) dir.mul(maxdist/dist);
+            dir.rotate_around_z(-camera1->yaw*RAD);
+            if(basenumbers)
+            {
+                static string blip;
+                formatstring(blip)("%d", i+1);
+                int tw, th;
+                text_bounds(blip, tw, th);
+                draw_text(blip, int(0.5f*(dir.x*fw/blipsize - tw)), int(0.5f*(dir.y*fh/blipsize - th)));
+            }
+            else
+            {
+                if(!blips) glBegin(GL_QUADS);
+                float x = 0.5f*(dir.x*fw/blipsize - fw), y = 0.5f*(dir.y*fh/blipsize - fh);
+                glTexCoord2f(0.0f, 0.0f); glVertex2f(x,    y);
+                glTexCoord2f(1.0f, 0.0f); glVertex2f(x+fw, y);
+                glTexCoord2f(1.0f, 1.0f); glVertex2f(x+fw, y+fh);
+                glTexCoord2f(0.0f, 1.0f); glVertex2f(x,    y+fh);
+            }
+            blips++;
         }
-        glEnd();
+        if(blips && !basenumbers) glEnd();
     }
 
     int respawnwait(fpsent *d)
@@ -437,23 +461,54 @@ struct captureclientmode : clientmode
 
     int clipconsole(int w, int h)
     {
-        return w*6/40;
+        return (h*(1 + 1 + 10))/(4*10);
     }
 
     void drawhud(fpsent *d, int w, int h)
     {
         glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
-        int x = 1800*w/h*34/40, y = 1800*1/40, s = 1800*w/h*5/40;
+        int s = 1800/4, x = 1800*w/h - s - s/10, y = s/10;
+        glColor4f(1, 1, 1, minimapalpha);
+        if(minimapalpha >= 1) glDisable(GL_BLEND);
+        bindminimap();
+        drawminimap(d, x, y, s);
+        if(minimapalpha >= 1) glEnable(GL_BLEND);
         glColor3f(1, 1, 1);
-        settexture("packages/hud/radar.png");
-        glBegin(GL_QUADS);
-        drawradar(float(x), float(y), float(s));
-        glEnd();
+        float margin = 0.04f, roffset = s*margin, rsize = s + 2*roffset;
+        settexture("packages/hud/radar.png", 3);
+        drawradar(x - roffset, y - roffset, rsize);
+        #if 0
+        settexture("packages/hud/compass.png", 3);
+        glPushMatrix();
+        glTranslatef(x - roffset + 0.5f*rsize, y - roffset + 0.5f*rsize, 0);
+        glRotatef(camera1->yaw + 180, 0, 0, -1);
+        drawradar(-0.5f*rsize, -0.5f*rsize, rsize);
+        glPopMatrix();
+        #endif
         bool showenemies = lastmillis%1000 >= 500;
-        drawblips(d, x, y, s, 1, showenemies);
-        drawblips(d, x, y, s, 0, showenemies);
-        drawblips(d, x, y, s, -1, showenemies);
-        if(showenemies) drawblips(d, x, y, s, -2);
+        int fw = 1, fh = 1;
+        if(basenumbers)
+        {
+            pushfont();
+            setfont("digit_blue");
+            text_bounds(" ", fw, fh);
+        }
+        else settexture("packages/hud/blip_blue.png", 3);
+        glPushMatrix();
+        glTranslatef(x + 0.5f*s, y + 0.5f*s, 0);
+        float blipsize = basenumbers ? 0.1f : 0.05f;
+        glScalef((s*blipsize)/fw, (s*blipsize)/fh, 1.0f);
+        drawblips(d, blipsize, fw, fh, 1, showenemies);
+        if(basenumbers) setfont("digit_grey");
+        else settexture("packages/hud/blip_grey.png", 3);
+        drawblips(d, blipsize, fw, fh, 0, showenemies);
+        if(basenumbers) setfont("digit_red");
+        else settexture("packages/hud/blip_red.png", 3);
+        drawblips(d, blipsize, fw, fh, -1, showenemies);
+        if(showenemies) drawblips(d, blipsize, fw, fh, -2);
+        glPopMatrix();
+        if(basenumbers) popfont();
+        drawteammates(d, x, y, s);
         if(d->state == CS_DEAD)
         {
             int wait = respawnwait(d);
@@ -479,23 +534,18 @@ struct captureclientmode : clientmode
             b.o = e->o;
             b.ammopos = b.o;
             abovemodel(b.ammopos, "base/neutral");
-            b.ammopos.z += FIREBALLRADIUS-2;
+            b.ammopos.z += AMMOHEIGHT-2;
             b.ammotype = e->attr1;
             defformatstring(alias)("base_%d", e->attr2);
             const char *name = getalias(alias);
-            if(name[0]) copystring(b.name, name); else formatstring(b.name)("base %d", bases.length());
+            copystring(b.name, name);
             b.light = e->light;
         }
-        vec center(0, 0, 0);
-        loopv(bases) center.add(bases[i].o);
-        center.div(bases.length());
-        radarscale = 0;
-        loopv(bases) radarscale = max(radarscale, 2*center.dist(bases[i].o));
     }
 
     void senditems(packetbuf &p)
     {
-        putint(p, SV_BASES);
+        putint(p, N_BASES);
         putint(p, bases.length());
         loopv(bases)
         {
@@ -515,36 +565,45 @@ struct captureclientmode : clientmode
         {
             if(strcmp(b.owner, owner))
             {
-                conoutf(CON_GAMEINFO, "%s captured %s", owner, b.name);
+                if(!b.name[0]) conoutf(CON_GAMEINFO, "%s captured base %d", teamcolor(owner, owner), i+1);
+                else if(basenumbers) conoutf(CON_GAMEINFO, "%s captured %s (%d)", teamcolor(owner, owner), b.name, i+1);
+                else conoutf(CON_GAMEINFO, "%s captured %s", teamcolor(owner, owner), b.name);
                 if(!strcmp(owner, player1->team)) playsound(S_V_BASECAP);
             }
         }
         else if(b.owner[0])
         {
-            conoutf(CON_GAMEINFO, "%s lost %s", b.owner, b.name);
+            if(!b.name[0]) conoutf(CON_GAMEINFO, "%s lost base %d", teamcolor(b.owner, b.owner), i+1);
+            else if(basenumbers) conoutf(CON_GAMEINFO, "%s lost %s (%d)", teamcolor(b.owner, b.owner), b.name, i+1);
+            else conoutf(CON_GAMEINFO, "%s lost %s", teamcolor(b.owner, b.owner), b.name);
             if(!strcmp(b.owner, player1->team)) playsound(S_V_BASELOST);
         }
-        if(strcmp(b.owner, owner)) particle_splash(PART_SPARK, 200, 250, b.ammopos, 0xB49B4B, 0.24f);
+        if(strcmp(b.owner, owner)) particle_splash(PART_SPARK, 200, 250, b.ammopos, owner[0] ? (strcmp(owner, player1->team) ? 0x802020 : 0x2020FF) : 0x208020, 0.24f);
         copystring(b.owner, owner);
         copystring(b.enemy, enemy);
         b.converted = converted;
-        if(ammo>b.ammo) playsound(S_ITEMSPAWN, &b.o);
+        if(ammo>b.ammo)
+        {
+            playsound(S_ITEMSPAWN, &b.o);
+            int icon = b.valid() ? itemstats[b.ammotype-1].icon : -1;
+            if(icon >= 0) particle_icon(vec(b.ammopos.x, b.ammopos.y, b.ammopos.z + AMMOHEIGHT + 1.0f), icon%4, icon/4, PART_HUD_ICON, 2000, 0xFFFFFF, 2.0f, -8);
+        }
         b.ammo = ammo;
     }
 
     void setscore(int base, const char *team, int total)
     {
         findscore(team).total = total;
-        if(total>=10000) conoutf(CON_GAMEINFO, "team %s captured all bases", team);
+        if(total>=10000) conoutf(CON_GAMEINFO, "%s captured all bases", teamcolor(team, team));
         else if(bases.inrange(base))
         {
             baseinfo &b = bases[base];
             if(!strcmp(b.owner, team))
             {
-                defformatstring(msg)("@%d", total);
+                defformatstring(msg)("%d", total);
                 vec above(b.ammopos);
-                above.z += FIREBALLRADIUS+1.0f;
-                particle_text(above, msg, PART_TEXT, 2000, isteam(team, player1->team) ? 0x6496FF : 0xFF4B19, 4.0f, -8);
+                above.z += AMMOHEIGHT+1.0f;
+                particle_textcopy(above, msg, PART_TEXT, 2000, isteam(team, player1->team) ? 0x6496FF : 0xFF4B19, 4.0f, -8);
             }
         }
     }
@@ -610,9 +669,6 @@ struct captureclientmode : clientmode
         findplayerspawn(d, pickteamspawn(d->team));
     }
 
-    const char *prefixnextmap() { return "capture_"; }
-
-
 	bool aicheck(fpsent *d, ai::aistate &b)
 	{
 		return false;
@@ -624,18 +680,19 @@ struct captureclientmode : clientmode
 		loopvj(bases)
 		{
 			baseinfo &f = bases[j];
+            if(!f.valid()) continue;
 			static vector<int> targets; // build a list of others who are interested in this
-			targets.setsizenodelete(0);
+			targets.setsize(0);
 			ai::checkothers(targets, d, ai::AI_S_DEFEND, ai::AI_T_AFFINITY, j, true);
 			fpsent *e = NULL;
 			int regen = !m_regencapture || d->health >= 100 ? 0 : 1;
 			if(m_regencapture)
 			{
 				int gun = f.ammotype-1+I_SHELLS;
-				if(f.ammo > 0 && f.ammotype > 0 && f.ammotype <= I_CARTRIDGES-I_SHELLS+1 && !d->hasmaxammo(gun))
+				if(f.ammo > 0 && !d->hasmaxammo(gun))
 					regen = gun != d->ai->weappref ? 2 : 4;
 			}
-			loopi(numdynents()) if((e = (fpsent *)iterdynents(i)) && ai::targetable(d, e, false) && !e->ai && d->team == e->team)
+			loopi(numdynents()) if((e = (fpsent *)iterdynents(i)) && !e->ai && e->state == CS_ALIVE && isteam(d->team, e->team))
 			{ // try to guess what non ai are doing
 				vec ep = e->feetpos();
 				if(targets.find(e->clientnum) < 0 && ep.squaredist(f.o) <= (CAPTURERADIUS*CAPTURERADIUS))
@@ -645,7 +702,7 @@ struct captureclientmode : clientmode
 			{
 				ai::interest &n = interests.add();
 				n.state = ai::AI_S_DEFEND;
-				n.node = ai::closestwaypoint(f.o, ai::NEARDIST, false);
+				n.node = ai::closestwaypoint(f.o, ai::SIGHTMIN, false);
 				n.target = j;
 				n.targtype = ai::AI_T_AFFINITY;
 				n.score = pos.squaredist(f.o)/(regen ? float(100*regen) : 1.f);
@@ -655,44 +712,42 @@ struct captureclientmode : clientmode
 
 	bool aidefend(fpsent *d, ai::aistate &b)
 	{
-		if(bases.inrange(b.target))
+        if(!bases.inrange(b.target)) return false;
+        baseinfo &f = bases[b.target];
+        if(!f.valid()) return false;
+		bool regen = !m_regencapture || d->health >= 100 ? false : true;
+		if(!regen && m_regencapture)
 		{
-			baseinfo &f = bases[b.target];
-			bool regen = !m_regencapture || d->health >= 100 ? false : true;
-			if(!regen && m_regencapture)
-			{
-				int gun = f.ammotype-1+I_SHELLS;
-				if(f.ammo > 0 && f.ammotype > 0 && f.ammotype <= I_CARTRIDGES-I_SHELLS+1 && !d->hasmaxammo(gun))
-					regen = true;
-			}
-			int walk = 0;
-			if(!regen && !f.enemy[0] && f.owner[0] && !strcmp(f.owner, d->team))
-			{
-				static vector<int> targets; // build a list of others who are interested in this
-				targets.setsizenodelete(0);
-				ai::checkothers(targets, d, ai::AI_S_DEFEND, ai::AI_T_AFFINITY, b.target, true);
-				fpsent *e = NULL;
-				loopi(numdynents()) if((e = (fpsent *)iterdynents(i)) && ai::targetable(d, e, false) && !e->ai && !strcmp(d->team, e->team))
-				{ // try to guess what non ai are doing
-					vec ep = e->feetpos();
-					if(targets.find(e->clientnum) < 0 && (ep.squaredist(f.o) <= (CAPTURERADIUS*CAPTURERADIUS*4)))
-						targets.add(e->clientnum);
-				}
-				if(!targets.empty())
-				{
-					if(lastmillis-b.millis >= (201-d->skill)*33)
-					{
-						d->ai->clear = true; // re-evaluate so as not to herd
-						return true;
-					}
-					else walk = 2;
-				}
-				else walk = 1;
-				b.millis = lastmillis;
-			}
-			return ai::defend(d, b, f.o, float(CAPTURERADIUS), float(CAPTURERADIUS*(2+(walk*2))), walk); // less wander than ctf
+			int gun = f.ammotype-1+I_SHELLS;
+			if(f.ammo > 0 && !d->hasmaxammo(gun))
+				regen = true;
 		}
-		return false;
+		int walk = 0;
+		if(!regen && !f.enemy[0] && f.owner[0] && !strcmp(f.owner, d->team))
+		{
+			static vector<int> targets; // build a list of others who are interested in this
+			targets.setsize(0);
+			ai::checkothers(targets, d, ai::AI_S_DEFEND, ai::AI_T_AFFINITY, b.target, true);
+			fpsent *e = NULL;
+			loopi(numdynents()) if((e = (fpsent *)iterdynents(i)) && !e->ai && e->state == CS_ALIVE && isteam(d->team, e->team))
+			{ // try to guess what non ai are doing
+				vec ep = e->feetpos();
+				if(targets.find(e->clientnum) < 0 && (ep.squaredist(f.o) <= (CAPTURERADIUS*CAPTURERADIUS*4)))
+					targets.add(e->clientnum);
+			}
+			if(!targets.empty())
+			{
+				if(lastmillis-b.millis >= (201-d->skill)*33)
+				{
+					d->ai->trywipe = true; // re-evaluate so as not to herd
+					return true;
+				}
+				else walk = 2;
+			}
+			else walk = 1;
+			b.millis = lastmillis;
+		}
+		return ai::defend(d, b, f.o, float(CAPTURERADIUS), float(CAPTURERADIUS*(2+(walk*2))), walk); // less wander than ctf
 	}
 
 	bool aipursue(fpsent *d, ai::aistate &b)
@@ -702,19 +757,62 @@ struct captureclientmode : clientmode
 	}
 };
 
+extern captureclientmode capturemode;
+ICOMMAND(repammo, "", (), capturemode.replenishammo());
+ICOMMAND(insidebases, "", (),
+{
+    vector<char> buf;
+    if(m_capture && player1->state == CS_ALIVE) loopv(capturemode.bases)
+    {
+        captureclientmode::baseinfo &b = capturemode.bases[i];
+        if(b.valid() && capturemode.insidebase(b, player1->feetpos()))
+        {
+            if(buf.length()) buf.add(' ');
+            defformatstring(basenum)("%d", i+1);
+            buf.put(basenum, strlen(basenum));
+        }
+    }
+    buf.add('\0');
+    result(buf.getbuf());
+}); 
+    
 #else
-    int scoresec;
     bool notgotbases;
 
-    captureservmode() : captures(0), scoresec(0), notgotbases(false) {}
+    captureservmode() : captures(0), notgotbases(false) {}
 
     void reset(bool empty)
     {
         resetbases();
-        scoresec = 0;
         notgotbases = !empty;
     }
 
+    void cleanup()
+    {
+        reset(false);
+    }
+
+    void setup()
+    {
+        reset(false);
+        if(notgotitems || ments.empty()) return;
+        loopv(ments)
+        {
+            entity &e = ments[i];
+            if(e.type != BASE) continue;
+            int ammotype = e.attr1;
+            addbase(ammotype, e.o); 
+        }
+        notgotbases = false;
+        sendbases();
+        loopv(clients) if(clients[i]->state.state==CS_ALIVE) entergame(clients[i]);
+    }
+
+    void newmap()
+    {
+        reset(true);
+    }
+
     void stealbase(int n, const char *team)
     {
         baseinfo &b = bases[n];
@@ -733,10 +831,10 @@ struct captureclientmode : clientmode
         loopv(bases)
         {
             baseinfo &b = bases[i];
-            if(b.ammotype>0 && b.ammotype<=I_CARTRIDGES-I_SHELLS+1 && insidebase(b, ci->state.o) && !ci->state.hasmaxammo(b.ammotype-1+I_SHELLS) && b.takeammo(ci->team))
+            if(b.valid() && insidebase(b, ci->state.o) && !ci->state.hasmaxammo(b.ammotype-1+I_SHELLS) && b.takeammo(ci->team))
             {
                 sendbaseinfo(i);
-                sendf(-1, 1, "riii", SV_REPAMMO, ci->clientnum, b.ammotype);
+                sendf(-1, 1, "riii", N_REPAMMO, ci->clientnum, b.ammotype);
                 ci->state.addammo(b.ammotype);
                 break;
             }
@@ -745,10 +843,11 @@ struct captureclientmode : clientmode
 
     void movebases(const char *team, const vec &oldpos, bool oldclip, const vec &newpos, bool newclip)
     {
-        if(!team[0] || minremain<=0) return;
+        if(!team[0] || gamemillis>=gamelimit) return;
         loopv(bases)
         {
             baseinfo &b = bases[i];
+            if(!b.valid()) continue;
             bool leave = !oldclip && insidebase(b, oldpos),
                  enter = !newclip && insidebase(b, newpos);
             if(leave && !enter && b.leave(team)) sendbaseinfo(i);
@@ -772,7 +871,7 @@ struct captureclientmode : clientmode
         if(!n) return;
         score &cs = findscore(team);
         cs.total += n;
-        sendf(-1, 1, "riisi", SV_BASESCORE, base, team, cs.total);
+        sendf(-1, 1, "riisi", N_BASESCORE, base, team, cs.total);
     }
 
     void regenowners(baseinfo &b, int ticks)
@@ -788,35 +887,41 @@ struct captureclientmode : clientmode
                     ci->state.health = min(ci->state.health + ticks*REGENHEALTH, ci->state.maxhealth);
                     notify = true;
                 }
-                if(ci->state.armour < itemstats[I_GREENARMOUR-I_SHELLS].max)
+                if(ci->state.armourtype != A_GREEN || ci->state.armour < itemstats[I_GREENARMOUR-I_SHELLS].max)
                 {
+                    if(ci->state.armourtype != A_GREEN)
+                    {
+                        ci->state.armourtype = A_GREEN;
+                        ci->state.armour = 0;
+                    }
                     ci->state.armour = min(ci->state.armour + ticks*REGENARMOUR, itemstats[I_GREENARMOUR-I_SHELLS].max);
                     notify = true;
                 }
-                if(b.ammotype>0)
+                if(b.valid())
                 {
                     int ammotype = b.ammotype-1+I_SHELLS;
-                    if(ammotype<=I_CARTRIDGES && !ci->state.hasmaxammo(ammotype))
+                    if(!ci->state.hasmaxammo(ammotype))
                     {
                         ci->state.addammo(b.ammotype, ticks*REGENAMMO, 100);
                         notify = true;
                     }
                 }
                 if(notify)
-                    sendf(-1, 1, "ri6", SV_BASEREGEN, ci->clientnum, ci->state.health, ci->state.armour, b.ammotype, b.ammotype>0 ? ci->state.ammo[b.ammotype] : 0);
+                    sendf(-1, 1, "ri6", N_BASEREGEN, ci->clientnum, ci->state.health, ci->state.armour, b.ammotype, b.valid() ? ci->state.ammo[b.ammotype] : 0);
             }
         }
     }
 
     void update()
     {
-        if(minremain<=0) return;
+        if(gamemillis>=gamelimit) return;
         endcheck();
         int t = gamemillis/1000 - (gamemillis-curtime)/1000;
         if(t<1) return;
         loopv(bases)
         {
             baseinfo &b = bases[i];
+            if(!b.valid()) continue;
             if(b.enemy[0])
             {
                 if(!b.owners || !b.enemies) b.occupy(b.enemy, OCCUPYBONUS*(b.enemies ? 1 : -1) + OCCUPYPOINTS*(b.enemies ? b.enemies : -(1+b.owners))*t);
@@ -846,7 +951,7 @@ struct captureclientmode : clientmode
     void sendbaseinfo(int i)
     {
         baseinfo &b = bases[i];
-        sendf(-1, 1, "riissii", SV_BASEINFO, i, b.owner, b.enemy, b.enemy[0] ? b.converted : 0, b.owner[0] ? b.ammo : 0);
+        sendf(-1, 1, "riissii", N_BASEINFO, i, b.owner, b.enemy, b.enemy[0] ? b.converted : 0, b.owner[0] ? b.ammo : 0);
     }
 
     void sendbases()
@@ -863,18 +968,18 @@ struct captureclientmode : clientmode
             loopv(scores)
             {
                 score &cs = scores[i];
-                putint(p, SV_BASESCORE);
+                putint(p, N_BASESCORE);
                 putint(p, -1);
                 sendstring(cs.team, p);
                 putint(p, cs.total);
             }
         }
-        putint(p, SV_BASES);
+        putint(p, N_BASES);
         putint(p, bases.length());
         loopv(bases)
         {
             baseinfo &b = bases[i];
-            putint(p, min(max(b.ammotype, 1), I_CARTRIDGES+1));
+            putint(p, b.ammotype);
             sendstring(b.owner, p);
             sendstring(b.enemy, p);
             putint(p, b.converted);
@@ -889,6 +994,7 @@ struct captureclientmode : clientmode
         loopv(bases)
         {
             baseinfo &b = bases[i];
+            if(!b.valid()) continue;
             if(b.owner[0])
             {
                 if(!lastteam) lastteam = b.owner;
@@ -907,7 +1013,7 @@ struct captureclientmode : clientmode
 
         if(!lastteam) return;
         findscore(lastteam).total = 10000;
-        sendf(-1, 1, "riisi", SV_BASESCORE, -1, lastteam, 10000);
+        sendf(-1, 1, "riisi", N_BASESCORE, -1, lastteam, 10000);
         startintermission();
     }
 
@@ -935,6 +1041,11 @@ struct captureclientmode : clientmode
         leavebases(ci->team, ci->state.o);
     }
 
+    bool canspawn(clientinfo *ci, bool connecting)
+    {
+        return m_regencapture || connecting || !ci->state.lastdeath || gamemillis+curtime-ci->state.lastdeath >= RESPAWNSECS*1000;
+    }
+
     void moved(clientinfo *ci, const vec &oldpos, bool oldclip, const vec &newpos, bool newclip)
     {
         if(notgotbases) return;
@@ -955,9 +1066,7 @@ struct captureclientmode : clientmode
         {
             int ammotype = getint(p);
             vec o;
-            o.x = getint(p)/DMF;
-            o.y = getint(p)/DMF;
-            o.z = getint(p)/DMF;
+            loopk(3) o[k] = max(getint(p)/DMF, 0.0f);
             if(p.overread()) break;
             if(commit && notgotbases) addbase(ammotype>=GUN_SG && ammotype<=GUN_PISTOL ? ammotype : min(ammotype, 0), o);
         }
@@ -983,17 +1092,17 @@ struct captureclientmode : clientmode
 
 #elif SERVMODE
 
-case SV_BASES:
+case N_BASES:
     if(smode==&capturemode) capturemode.parsebases(p, (ci->state.state!=CS_SPECTATOR || ci->privilege || ci->local) && !strcmp(ci->clientmap, smapname));
     break;
 
-case SV_REPAMMO:
-    if(ci->state.state!=CS_SPECTATOR && cq && smode==&capturemode) capturemode.replenishammo(cq);
+case N_REPAMMO:
+    if((ci->state.state!=CS_SPECTATOR || ci->local || ci->privilege) && cq && smode==&capturemode) capturemode.replenishammo(cq);
     break;
 
 #else
 
-case SV_BASEINFO:
+case N_BASEINFO:
 {
     int base = getint(p);
     string owner, enemy;
@@ -1006,20 +1115,21 @@ case SV_BASEINFO:
     break;
 }
 
-case SV_BASEREGEN:
+case N_BASEREGEN:
 {
     int rcn = getint(p), health = getint(p), armour = getint(p), ammotype = getint(p), ammo = getint(p);
     fpsent *regen = rcn==player1->clientnum ? player1 : getclient(rcn);
     if(regen && m_capture)
     {
         regen->health = health;
+        regen->armourtype = A_GREEN;
         regen->armour = armour;
         if(ammotype>=GUN_SG && ammotype<=GUN_PISTOL) regen->ammo[ammotype] = ammo;
     }
     break;
 }
 
-case SV_BASES:
+case N_BASES:
 {
     int numbases = getint(p);
     loopi(numbases)
@@ -1036,7 +1146,7 @@ case SV_BASES:
     break;
 }
 
-case SV_BASESCORE:
+case N_BASESCORE:
 {
     int base = getint(p);
     getstring(text, p);
@@ -1045,7 +1155,7 @@ case SV_BASESCORE:
     break;
 }
 
-case SV_REPAMMO:
+case N_REPAMMO:
 {
     int rcn = getint(p), ammotype = getint(p);
     fpsent *r = rcn==player1->clientnum ? player1 : getclient(rcn);
