@@ -1,6 +1,6 @@
 
 #define EXT_ACK                         -1
-#define EXT_VERSION                     104
+#define EXT_VERSION                     105
 #define EXT_NO_ERROR                    0
 #define EXT_ERROR                       1
 #define EXT_PLAYERSTATS_RESP_IDS        -10
@@ -52,47 +52,33 @@
         sendserverinforeply(q);
     }
 
+    static inline void extinfoteamscore(ucharbuf &p, const char *team, int score)
+    {
+        sendstring(team, p);
+        putint(p, score);
+        if(!smode || !smode->extinfoteam(team, p))
+            putint(p,-1); //no bases follow
+    }
+
     void extinfoteams(ucharbuf &p)
     {
         putint(p, m_teammode ? 0 : 1);
         putint(p, gamemode);
-        putint(p, minremain);
+        putint(p, max((gamelimit - gamemillis)/1000, 0));
         if(!m_teammode) return;
 
         vector<teamscore> scores;
-
-        //most taken from scoreboard.h
-        if(smode && smode->hidefrags()) 
+        if(smode && smode->hidefrags()) smode->getteamscores(scores);
+        loopv(clients)
         {
-            smode->getteamscores(scores);
-            loopv(clients) if(clients[i]->team[0])
+            clientinfo *ci = clients[i];
+            if(ci->state.state!=CS_SPECTATOR && ci->team[0] && scores.htfind(ci->team) < 0)
             {
-                clientinfo *ci = clients[i];
-                teamscore *ts = NULL;
-                loopvj(scores) if(!strcmp(scores[j].team, ci->team)) { ts = &scores[j]; break; }
-                if(!ts) scores.add(teamscore(ci->team, 0));
+                if(smode && smode->hidefrags()) scores.add(teamscore(ci->team, 0));
+                else { teaminfo *ti = teaminfos.access(ci->team); scores.add(teamscore(ci->team, ti ? ti->frags : 0)); }
             }
         }
-        else
-        {
-            loopv(clients) if(clients[i]->team[0])
-            {
-                clientinfo *ci = clients[i];
-                teamscore *ts = NULL;
-                loopvj(scores) if(!strcmp(scores[j].team, ci->team)) { ts = &scores[j]; break; }
-                if(!ts) scores.add(teamscore(ci->team, ci->state.frags));
-                else ts->score += ci->state.frags;
-            }
-        }
-
-        loopv(scores)
-        {
-            sendstring(scores[i].team, p);
-            putint(p, scores[i].score);
-
-            if(!smode || !smode->extinfoteam(scores[i].team, p))
-                putint(p,-1); //no bases follow
-        }
+        loopv(scores) extinfoteamscore(p, scores[i].team, scores[i].score);
     }
 
     void extserverinforeply(ucharbuf &req, ucharbuf &p)
@@ -107,7 +93,7 @@
         {
             case EXT_UPTIME:
             {
-                putint(p, uint(totalmillis)/1000); //in seconds
+                putint(p, totalsecs); //in seconds
                 break;
             }
 
