@@ -3,7 +3,7 @@
 namespace game
 {
     bool intermission = false;
-    int maptime = 0, maprealtime = 0, minremain = 0;
+    int maptime = 0, maprealtime = 0, maplimit = -1;
     int respawnent = -1;
     int lasthit = 0, lastspawnattempt = 0;
 
@@ -20,10 +20,16 @@ namespace game
         if(player1->state!=CS_ALIVE || player1->physstate<PHYS_SLOPE) return;
         if(lastmillis-player1->lasttaunt<1000) return;
         player1->lasttaunt = lastmillis;
-        addmsg(SV_TAUNT, "rc", player1);
+        addmsg(N_TAUNT, "rc", player1);
     }
     COMMAND(taunt, "");
 
+    ICOMMAND(getfollow, "", (),
+    {
+        fpsent *f = followingplayer();
+        intret(f ? f->clientnum : -1);
+    });
+
 	void follow(char *arg)
     {
         if(arg[0] ? player1->state==CS_SPECTATOR : following>=0)
@@ -68,7 +74,7 @@ namespace game
         {
             clearmovables();
             clearmonsters();                 // all monsters back at their spawns for editing
-            resettriggers();
+            entities::resettriggers();
         }
         clearprojectiles();
         clearbouncers();
@@ -83,14 +89,11 @@ namespace game
 
     void respawnself()
     {
-        if(paused || ispaused()) return;
+        if(ispaused()) return;
         if(m_mp(gamemode))
         {
-            if(player1->respawned!=player1->lifesequence)
-            {
-                addmsg(SV_TRYSPAWN, "rc", player1);
-                player1->respawned = player1->lifesequence;
-            }
+            int seq = (player1->lifesequence<<16)|((lastmillis/1000)&0xFFFF);
+            if(player1->respawned!=seq) { addmsg(N_TRYSPAWN, "rc", player1); player1->respawned = seq; }
         }
         else
         {
@@ -166,6 +169,7 @@ namespace game
         d->o = d->newpos;
         d->yaw = d->newyaw;
         d->pitch = d->newpitch;
+        d->roll = d->newroll;
         if(move)
         {
             moveplayer(d, 1, false);
@@ -179,6 +183,7 @@ namespace game
             if(d->yaw<0) d->yaw += 360;
             else if(d->yaw>=360) d->yaw -= 360;
             d->pitch += d->deltapitch*k;
+            d->roll += d->deltaroll*k;
         }
     }
 
@@ -189,14 +194,14 @@ namespace game
             fpsent *d = players[i];
             if(d == player1 || d->ai) continue;
 
-            if(d->state==CS_ALIVE)
+            if(d->state==CS_DEAD && d->ragdoll) moveragdoll(d);
+            else if(!intermission)
             {
                 if(lastmillis - d->lastaction >= d->gunwait) d->gunwait = 0;
                 if(d->quadmillis) entities::checkquad(curtime, d);
             }
-            else if(d->state==CS_DEAD && d->ragdoll) moveragdoll(d);
 
-            const int lagtime = lastmillis-d->lastupdate;
+            const int lagtime = totalmillis-d->lastupdate;
             if(!lagtime || intermission) continue;
             else if(lagtime>1000 && d->state==CS_ALIVE)
             {
@@ -212,15 +217,12 @@ namespace game
         }
     }
 
-    VARFP(slowmosp, 0, 0, 1,
-    {
-        if(m_sp && !slowmosp) setvar("gamespeed", 100);
-    });
+    VARFP(slowmosp, 0, 0, 1, { if(m_sp && !slowmosp) server::forcegamespeed(100); }); 
 
     void checkslowmo()
     {
         static int lastslowmohealth = 0;
-        setvar("gamespeed", intermission ? 100 : clamp(player1->health, 25, 200), true, false);
+        server::forcegamespeed(intermission ? 100 : clamp(player1->health, 25, 200));
         if(player1->health<player1->maxhealth && lastmillis-max(maptime, lastslowmohealth)>player1->health*player1->health/2)
         {
             lastslowmohealth = lastmillis;
@@ -234,8 +236,11 @@ namespace game
         if(!curtime) { gets2c(); if(player1->clientnum>=0) c2sinfo(); return; }
 
         physicsframe();
-        ai::trydropwaypoints();
-        entities::checkquad(curtime, player1);
+        ai::navigate();
+        if(player1->state != CS_DEAD && !intermission)
+        {
+            if(player1->quadmillis) entities::checkquad(curtime, player1);
+        }
         updateweapons(curtime);
         otherplayers(curtime);
         ai::update();
@@ -243,13 +248,13 @@ namespace game
         gets2c();
         updatemovables(curtime);
         updatemonsters(curtime);
-        if(player1->state==CS_DEAD)
+        if(player1->state == CS_DEAD)
         {
             if(player1->ragdoll) moveragdoll(player1);
             else if(lastmillis-player1->lastpain<2000)
             {
                 player1->move = player1->strafe = 0;
-                moveplayer(player1, 10, false);
+                moveplayer(player1, 10, true);
             }
         }
         else if(!intermission)
@@ -261,7 +266,7 @@ namespace game
             if(m_sp)
             {
                 if(slowmosp) checkslowmo();
-                if(m_classicsp) checktriggers();
+                if(m_classicsp) entities::checktriggers();
             }
             else if(cmode) cmode->checkitems(player1);
         }
@@ -330,11 +335,11 @@ namespace game
 
     void damaged(int damage, fpsent *d, fpsent *actor, bool local)
     {
-        if(d->state!=CS_ALIVE || intermission) return;
+        if((d->state!=CS_ALIVE && d->state != CS_LAGGED && d->state != CS_SPAWNING) || intermission) return;
 
         if(local) damage = d->dodamage(damage);
         else if(actor==player1) return;
-  
+
         fpsent *h = hudplayer();
         if(h!=player1 && actor==h && d!=actor)
         {
@@ -357,14 +362,16 @@ namespace game
         else playsound(S_PAIN1+rnd(5), &d->o);
     }
 
+    VARP(deathscore, 0, 1, 1);
+
     void deathstate(fpsent *d, bool restore)
     {
         d->state = CS_DEAD;
         d->lastpain = lastmillis;
-        d->superdamage = restore ? 0 : max(-d->health, 0);
+        if(!restore) gibeffect(max(-d->health, 0), d->vel, d);
         if(d==player1)
         {
-            showscores(true);
+            if(deathscore) showscores(true);
             disablezoom();
             if(!restore) loopi(NUMGUNS) savedammo[i] = player1->ammo[i];
             d->attacking = false;
@@ -379,10 +386,11 @@ namespace game
             d->resetinterp();
             d->smoothmillis = 0;
             playsound(S_DIE1+rnd(2), &d->o);
-            if(!restore) superdamageeffect(d->vel, d);
         }
     }
 
+    VARP(teamcolorfrags, 0, 1, 1);
+
     void killed(fpsent *d, fpsent *actor)
     {
         if(d->state==CS_EDITING)
@@ -392,57 +400,76 @@ namespace game
             else d->resetinterp();
             return;
         }
-        else if(d->state!=CS_ALIVE || intermission) return;
+        else if((d->state!=CS_ALIVE && d->state != CS_LAGGED && d->state != CS_SPAWNING) || intermission) return;
 
         fpsent *h = followingplayer();
         if(!h) h = player1;
         int contype = d==h || actor==h ? CON_FRAG_SELF : CON_FRAG_OTHER;
-        string dname, aname;
-        copystring(dname, d==player1 ? "you" : colorname(d));
-        copystring(aname, actor==player1 ? "you" : colorname(actor));
+        const char *dname = "", *aname = "";
+        if(m_teammode && teamcolorfrags)
+        {
+            dname = teamcolorname(d, "you");
+            aname = teamcolorname(actor, "you");
+        }
+        else
+        {
+            dname = colorname(d, NULL, "", "", "you");
+            aname = colorname(actor, NULL, "", "", "you");
+        }
         if(actor->type==ENT_AI)
             conoutf(contype, "\f2%s got killed by %s!", dname, aname);
         else if(d==actor || actor->type==ENT_INANIMATE)
             conoutf(contype, "\f2%s suicided%s", dname, d==player1 ? "!" : "");
         else if(isteam(d->team, actor->team))
         {
-            if(actor==player1) conoutf(contype, "\f3you fragged a teammate (%s)", dname);
-            else if(d==player1) conoutf(contype, "\f3you got fragged by a teammate (%s)", aname);
+            contype |= CON_TEAMKILL;
+            if(actor==player1) conoutf(contype, "\f6%s fragged a teammate (%s)", aname, dname);
+            else if(d==player1) conoutf(contype, "\f6%s got fragged by a teammate (%s)", dname, aname);
             else conoutf(contype, "\f2%s fragged a teammate (%s)", aname, dname);
         }
         else
         {
-            if(d==player1) conoutf(contype, "\f2you got fragged by %s", aname);
+            if(d==player1) conoutf(contype, "\f2%s got fragged by %s", dname, aname);
             else conoutf(contype, "\f2%s fragged %s", aname, dname);
         }
         deathstate(d);
 		ai::killed(d, actor);
     }
 
-    void timeupdate(int timeremain)
+    void timeupdate(int secs)
     {
-        minremain = timeremain;
-        if(!timeremain)
+        if(secs > 0)
+        {
+            maplimit = lastmillis + secs*1000;
+        }
+        else
         {
             intermission = true;
             player1->attacking = false;
             if(cmode) cmode->gameover();
             conoutf(CON_GAMEINFO, "\f2intermission:");
             conoutf(CON_GAMEINFO, "\f2game has ended!");
-            conoutf(CON_GAMEINFO, "\f2player frags: %d, deaths: %d", player1->frags, player1->deaths);
-            int accuracy = player1->totaldamage*100/max(player1->totalshots, 1);
+            if(m_ctf) conoutf(CON_GAMEINFO, "\f2player frags: %d, flags: %d, deaths: %d", player1->frags, player1->flags, player1->deaths);
+            else if(m_collect) conoutf(CON_GAMEINFO, "\f2player frags: %d, skulls: %d, deaths: %d", player1->frags, player1->flags, player1->deaths);
+            else conoutf(CON_GAMEINFO, "\f2player frags: %d, deaths: %d", player1->frags, player1->deaths);
+            int accuracy = (player1->totaldamage*100)/max(player1->totalshots, 1);
             conoutf(CON_GAMEINFO, "\f2player total damage dealt: %d, damage wasted: %d, accuracy(%%): %d", player1->totaldamage, player1->totalshots-player1->totaldamage, accuracy);
             if(m_sp) spsummary(accuracy);
 
             showscores(true);
             disablezoom();
-        }
-        else if(timeremain > 0)
-        {
-            conoutf(CON_GAMEINFO, "\f2time remaining: %d %s", timeremain, timeremain==1 ? "minute" : "minutes");
+            
+            if(identexists("intermission")) execute("intermission");
         }
     }
 
+    ICOMMAND(getfrags, "", (), intret(player1->frags));
+    ICOMMAND(getflags, "", (), intret(player1->flags));
+    ICOMMAND(getdeaths, "", (), intret(player1->deaths));
+    ICOMMAND(getaccuracy, "", (), intret((player1->totaldamage*100)/max(player1->totalshots, 1)));
+    ICOMMAND(gettotaldamage, "", (), intret(player1->totaldamage));
+    ICOMMAND(gettotalshots, "", (), intret(player1->totalshots));
+
     vector<fpsent *> clients;
 
     fpsent *newclient(int cn)   // ensure valid entity
@@ -480,9 +507,10 @@ namespace game
             if(followdir) nextfollow(followdir);
             else stopfollowing();
         }
+        unignore(cn);
         fpsent *d = clients[cn];
         if(!d) return;
-        if(notify && d->name[0]) conoutf("player %s disconnected", colorname(d));
+        if(notify && d->name[0]) conoutf("\f4leave:\f7 %s", colorname(d));
         removeweapons(d);
         removetrackedparticles(d);
         removetrackeddynlights(d);
@@ -507,7 +535,6 @@ namespace game
 
     void startgame()
     {
-        respawnent = -1;
         clearmovables();
         clearmonsters();
 
@@ -515,11 +542,13 @@ namespace game
         clearbouncers();
         clearragdolls();
 
+        clearteaminfo();
+
         // reset perma-state
         loopv(players)
         {
             fpsent *d = players[i];
-            d->frags = 0;
+            d->frags = d->flags = 0;
             d->deaths = 0;
             d->totaldamage = 0;
             d->totalshots = 0;
@@ -531,9 +560,10 @@ namespace game
         setclientmode();
 
         intermission = false;
-        maptime = 0;
+        maptime = maprealtime = 0;
+        maplimit = -1;
 
-        if(cmode) 
+        if(cmode)
         {
             cmode->preload();
             cmode->setup();
@@ -567,11 +597,12 @@ namespace game
         ai::savewaypoints();
         ai::clearwaypoints(true);
 
+        respawnent = -1; // so we don't respawn at an old spot
         if(!m_mp(gamemode)) spawnplayer(player1);
         else findplayerspawn(player1, -1);
         entities::resetspawns();
         copystring(clientmap, name ? name : "");
-
+        
         sendmapinfo();
     }
 
@@ -589,17 +620,26 @@ namespace game
         else if(floorlevel<0) { if(d==player1 || d->type!=ENT_PLAYER || ((fpsent *)d)->ai) msgsound(S_LAND, d); }
     }
 
+    void dynentcollide(physent *d, physent *o, const vec &dir)
+    {
+        switch(d->type)
+        {
+            case ENT_AI: if(dir.z > 0) stackmonster((monster *)d, o); break;
+            case ENT_INANIMATE: if(dir.z > 0) stackmovable((movable *)d, o); break;
+        }
+    }
+
     void msgsound(int n, physent *d)
     {
         if(!d || d==player1)
         {
-            addmsg(SV_SOUND, "ci", d, n);
+            addmsg(N_SOUND, "ci", d, n);
             playsound(n);
         }
-        else 
+        else
         {
             if(d->type==ENT_PLAYER && ((fpsent *)d)->ai)
-                addmsg(SV_SOUND, "ci", d, n);
+                addmsg(N_SOUND, "ci", d, n);
             playsound(n, &d->o);
         }
     }
@@ -616,22 +656,50 @@ namespace game
         return NULL;
     }
 
-    bool duplicatename(fpsent *d, const char *name = NULL)
+    bool duplicatename(fpsent *d, const char *name = NULL, const char *alt = NULL)
     {
         if(!name) name = d->name;
+        if(alt && d != player1 && !strcmp(name, alt)) return true;
         loopv(players) if(d!=players[i] && !strcmp(name, players[i]->name)) return true;
         return false;
     }
 
-    const char *colorname(fpsent *d, const char *name, const char *prefix)
+    static string cname[3];
+    static int cidx = 0;
+
+    const char *colorname(fpsent *d, const char *name, const char *prefix, const char *suffix, const char *alt)
     {
-        if(!name) name = d->name;
-        if(name[0] && !duplicatename(d, name) && d->aitype == AI_NONE) return name;
-        static string cname[3];
-        static int cidx = 0;
+        if(!name) name = alt && d == player1 ? alt : d->name; 
+        bool dup = !name[0] || duplicatename(d, name, alt) || d->aitype != AI_NONE;
+        if(dup || prefix[0] || suffix[0])
+        {
+            cidx = (cidx+1)%3;
+            if(dup) formatstring(cname[cidx])(d->aitype == AI_NONE ? "%s%s \fs\f5(%d)\fr%s" : "%s%s \fs\f5[%d]\fr%s", prefix, name, d->clientnum, suffix);
+            else formatstring(cname[cidx])("%s%s%s", prefix, name, suffix);
+            return cname[cidx];
+        }
+        return name;
+    }
+
+    VARP(teamcolortext, 0, 1, 1);
+
+    const char *teamcolorname(fpsent *d, const char *alt)
+    {
+        if(!teamcolortext || !m_teammode) return colorname(d, NULL, "", "", alt);
+        return colorname(d, NULL, isteam(d->team, player1->team) ? "\fs\f1" : "\fs\f3", "\fr", alt); 
+    }
+
+    const char *teamcolor(const char *name, bool sameteam, const char *alt)
+    {
+        if(!teamcolortext || !m_teammode) return sameteam || !alt ? name : alt;
         cidx = (cidx+1)%3;
-        formatstring(cname[cidx])(d->aitype == AI_NONE ? "%s%s \fs\f5(%d)\fr" : "%s%s \fs\f5[%d]\fr", prefix, name, d->clientnum);
+        formatstring(cname[cidx])(sameteam ? "\fs\f1%s\fr" : "\fs\f3%s\fr", sameteam || !alt ? name : alt);
         return cname[cidx];
+    }    
+    
+    const char *teamcolor(const char *name, const char *team, const char *alt)
+    {
+        return teamcolor(name, team && isteam(team, player1->team), alt);
     }
 
     void suicide(physent *d)
@@ -641,10 +709,10 @@ namespace game
             if(d->state!=CS_ALIVE) return;
             fpsent *pl = (fpsent *)d;
             if(!m_mp(gamemode)) killed(pl, pl);
-            else if(pl->suicided!=pl->lifesequence)
+            else 
             {
-                addmsg(SV_SUICIDE, "rc", pl);
-                pl->suicided = pl->lifesequence;
+                int seq = (pl->lifesequence<<16)|((lastmillis/1000)&0xFFFF);
+                if(pl->suicided!=seq) { addmsg(N_SUICIDE, "rc", pl); pl->suicided = seq; }
             }
         }
         else if(d->type==ENT_AI) suicidemonster((monster *)d);
@@ -652,19 +720,21 @@ namespace game
     }
     ICOMMAND(kill, "", (), suicide(player1));
 
+    bool needminimap() { return m_ctf || m_protect || m_hold || m_capture || m_collect; }
+
     void drawicon(int icon, float x, float y, float sz)
     {
         settexture("packages/hud/items.png");
-        glBegin(GL_QUADS);
+        glBegin(GL_TRIANGLE_STRIP);
         float tsz = 0.25f, tx = tsz*(icon%4), ty = tsz*(icon/4);
         glTexCoord2f(tx,     ty);     glVertex2f(x,    y);
         glTexCoord2f(tx+tsz, ty);     glVertex2f(x+sz, y);
-        glTexCoord2f(tx+tsz, ty+tsz); glVertex2f(x+sz, y+sz);
         glTexCoord2f(tx,     ty+tsz); glVertex2f(x,    y+sz);
+        glTexCoord2f(tx+tsz, ty+tsz); glVertex2f(x+sz, y+sz);
         glEnd();
     }
 
-    float abovegameplayhud()
+    float abovegameplayhud(int w, int h)
     {
         switch(hudplayer()->state)
         {
@@ -680,35 +750,19 @@ namespace game
         ammohuddown[3] = { GUN_RIFLE, GUN_SG, GUN_PISTOL },
         ammohudcycle[7] = { -1, -1, -1, -1, -1, -1, -1 };
 
-    ICOMMAND(ammohudup, "sss", (char *w1, char *w2, char *w3),
+    ICOMMAND(ammohudup, "V", (tagval *args, int numargs),
     {
-        int i = 0;
-        if(w1[0]) ammohudup[i++] = atoi(w1);
-        if(w2[0]) ammohudup[i++] = atoi(w2);
-        if(w3[0]) ammohudup[i++] = atoi(w3);
-        while(i < 3) ammohudup[i++] = -1;
+        loopi(3) ammohudup[i] = i < numargs ? getweapon(args[i].getstr()) : -1;
     });
 
-    ICOMMAND(ammohuddown, "sss", (char *w1, char *w2, char *w3),
+    ICOMMAND(ammohuddown, "V", (tagval *args, int numargs),
     {
-        int i = 0;
-        if(w1[0]) ammohuddown[i++] = atoi(w1);
-        if(w2[0]) ammohuddown[i++] = atoi(w2);
-        if(w3[0]) ammohuddown[i++] = atoi(w3);
-        while(i < 3) ammohuddown[i++] = -1;
+        loopi(3) ammohuddown[i] = i < numargs ? getweapon(args[i].getstr()) : -1;
     });
 
-    ICOMMAND(ammohudcycle, "sssssss", (char *w1, char *w2, char *w3, char *w4, char *w5, char *w6, char *w7),
+    ICOMMAND(ammohudcycle, "V", (tagval *args, int numargs),
     {
-        int i = 0;
-        if(w1[0]) ammohudcycle[i++] = atoi(w1);
-        if(w2[0]) ammohudcycle[i++] = atoi(w2);
-        if(w3[0]) ammohudcycle[i++] = atoi(w3);
-        if(w4[0]) ammohudcycle[i++] = atoi(w4);
-        if(w5[0]) ammohudcycle[i++] = atoi(w5);
-        if(w6[0]) ammohudcycle[i++] = atoi(w6);
-        if(w7[0]) ammohudcycle[i++] = atoi(w7);
-        while(i < 7) ammohudcycle[i++] = -1;
+        loopi(7) ammohudcycle[i] = i < numargs ? getweapon(args[i].getstr()) : -1;
     });
 
     VARP(ammohud, 0, 1, 1);
@@ -792,7 +846,16 @@ namespace game
             text_bounds(f ? colorname(f) : " ", fw, fh);
             fh = max(fh, ph);
             draw_text("SPECTATOR", w*1800/h - tw - pw, 1650 - th - fh);
-            if(f) draw_text(colorname(f), w*1800/h - fw - pw, 1650 - fh);
+            if(f) 
+            {
+                int color = f->state!=CS_DEAD ? 0xFFFFFF : 0x606060;
+                if(f->privilege)
+                {
+                    color = f->privilege>=PRIV_ADMIN ? 0xFF8000 : 0x40FF80;
+                    if(f->state==CS_DEAD) color = (color>>1)&0x7F7F7F;
+                }
+                draw_text(colorname(f), w*1800/h - fw - pw, 1650 - fh, (color>>16)&0xFF, (color>>8)&0xFF, color&0xFF);
+            }
         }
 
         fpsent *d = hudplayer();
@@ -866,11 +929,11 @@ namespace game
 
     bool serverinfostartcolumn(g3d_gui *g, int i)
     {
-        static const char *names[] = { "ping ", "players ", "map ", "mode ", "master ", "host ", "port ", "description " };
-        static const int struts[] =  { 0,       0,          12,     12,      8,         13,      6,       24 };
+        static const char *names[] =  { "ping ", "players ", "mode ", "map ", "time ", "master ", "host ", "port ", "description " };
+        static const float struts[] = { 7,       7,          12.5f,   14,      7,      8,         14,      7,       24.5f };
         if(size_t(i) >= sizeof(names)/sizeof(names[0])) return false;
         g->pushlist();
-        g->text(names[i], 0xFFFF80, !i ? "server" : NULL);
+        g->text(names[i], 0xFFFF80, !i ? " " : NULL);
         if(struts[i]) g->strut(struts[i]);
         g->mergehits(true);
         return true;
@@ -879,9 +942,20 @@ namespace game
     void serverinfoendcolumn(g3d_gui *g, int i)
     {
         g->mergehits(false);
+        g->column(i);
         g->poplist();
     }
 
+    const char *mastermodecolor(int n, const char *unknown)
+    {
+        return (n>=MM_START && size_t(n-MM_START)<sizeof(mastermodecolors)/sizeof(mastermodecolors[0])) ? mastermodecolors[n-MM_START] : unknown;
+    }
+
+    const char *mastermodeicon(int n, const char *unknown)
+    {
+        return (n>=MM_START && size_t(n-MM_START)<sizeof(mastermodeicons)/sizeof(mastermodeicons[0])) ? mastermodeicons[n-MM_START] : unknown;
+    }
+
     bool serverinfoentry(g3d_gui *g, int i, const char *name, int port, const char *sdesc, const char *map, int ping, const vector<int> &attr, int np)
     {
         if(ping < 0 || attr.empty() || attr[0]!=PROTOCOL_VERSION)
@@ -889,25 +963,26 @@ namespace game
             switch(i)
             {
                 case 0:
-                    if(g->button(" ", 0xFFFFDD, "server")&G3D_UP) return true;
+                    if(g->button(" ", 0xFFFFDD, "serverunk")&G3D_UP) return true;
                     break;
 
                 case 1:
                 case 2:
                 case 3:
                 case 4:
+                case 5:
                     if(g->button(" ", 0xFFFFDD)&G3D_UP) return true;
                     break;
 
-                case 5:
+                case 6:
                     if(g->buttonf("%s ", 0xFFFFDD, NULL, name)&G3D_UP) return true;
                     break;
 
-                case 6:
+                case 7:
                     if(g->buttonf("%d ", 0xFFFFDD, NULL, port)&G3D_UP) return true;
                     break;
 
-                case 7:
+                case 8:
                     if(ping < 0)
                     {
                         if(g->button(sdesc, 0xFFFFDD)&G3D_UP) return true;
@@ -921,38 +996,51 @@ namespace game
         switch(i)
         {
             case 0:
-                if(g->buttonf("%d ", 0xFFFFDD, "server", ping)&G3D_UP) return true;
+            {
+                const char *icon = attr.inrange(3) && np >= attr[3] ? "serverfull" : (attr.inrange(4) ? mastermodeicon(attr[4], "serverunk") : "serverunk");
+                if(g->buttonf("%d ", 0xFFFFDD, icon, ping)&G3D_UP) return true;
                 break;
+            }
 
             case 1:
                 if(attr.length()>=4)
                 {
-                    if(g->buttonf("%d/%d ", 0xFFFFDD, NULL, np, attr[3])&G3D_UP) return true;
+                    if(g->buttonf(np >= attr[3] ? "\f3%d/%d " : "%d/%d ", 0xFFFFDD, NULL, np, attr[3])&G3D_UP) return true;
                 }
                 else if(g->buttonf("%d ", 0xFFFFDD, NULL, np)&G3D_UP) return true;
                 break;
 
             case 2:
-                if(g->buttonf("%.25s ", 0xFFFFDD, NULL, map)&G3D_UP) return true;
-                break;
-
-            case 3:
                 if(g->buttonf("%s ", 0xFFFFDD, NULL, attr.length()>=2 ? server::modename(attr[1], "") : "")&G3D_UP) return true;
                 break;
 
-            case 4:
-                if(g->buttonf("%s ", 0xFFFFDD, NULL, attr.length()>=5 ? server::mastermodename(attr[4], "") : "")&G3D_UP) return true;
+            case 3:
+                if(g->buttonf("%.25s ", 0xFFFFDD, NULL, map)&G3D_UP) return true;
                 break;
 
+            case 4:
+                if(attr.length()>=3 && attr[2] > 0)
+                {
+                    int secs = clamp(attr[2], 0, 59*60+59),
+                        mins = secs/60;
+                    secs %= 60;
+                    if(g->buttonf("%d:%02d ", 0xFFFFDD, NULL, mins, secs)&G3D_UP) return true;
+                }
+                else if(g->buttonf(" ", 0xFFFFDD)&G3D_UP) return true;
+                break;
             case 5:
-                if(g->buttonf("%s ", 0xFFFFDD, NULL, name)&G3D_UP) return true;
+                if(g->buttonf("%s%s ", 0xFFFFDD, NULL, attr.length()>=5 ? mastermodecolor(attr[4], "") : "", attr.length()>=5 ? server::mastermodename(attr[4], "") : "")&G3D_UP) return true;
                 break;
 
             case 6:
-                if(g->buttonf("%d ", 0xFFFFDD, NULL, port)&G3D_UP) return true;
+                if(g->buttonf("%s ", 0xFFFFDD, NULL, name)&G3D_UP) return true;
                 break;
 
             case 7:
+                if(g->buttonf("%d ", 0xFFFFDD, NULL, port)&G3D_UP) return true;
+                break;
+
+            case 8:
                 if(g->buttonf("%.25s", 0xFFFFDD, NULL, sdesc)&G3D_UP) return true;
                 break;
         }
@@ -963,8 +1051,8 @@ namespace game
     void writegamedata(vector<char> &extras) {}
     void readgamedata(vector<char> &extras) {}
 
-    const char *gameident() { return "fps"; }
     const char *savedconfig() { return "config.cfg"; }
+    const char *restoreconfig() { return "restore.cfg"; }
     const char *defaultconfig() { return "data/defaults.cfg"; }
     const char *autoexec() { return "autoexec.cfg"; }
     const char *savedservers() { return "servers.cfg"; }
